# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.registry import RegistryManagerException
from anyblok_pyramid_rest_api.crud_resource import saved_errors_in_request


class TestBlokApi2:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_2):
        transaction = registry_rest_api_2.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_predicate(self, registry_rest_api_2, webserver):
        webserver.get('/bloks1', status=200)
        webserver.get('/bloks2', status=200)

    def test_multi_primary_keys(self, webserver):
        resp = webserver.get(
            '/column/Model.System.Model/name', status=200)
        assert resp.json_body == {
            'autoincrement': False,
            'code': 'system_model.name',
            'entity_type': 'Model.System.Column',
            'foreign_key': None,
            'ftype': 'String',
            'label': 'Name',
            'model': 'Model.System.Model',
            'name': 'name',
            'nullable': False,
            'primary_key': True,
            'remote_model': None,
            'unique': None
        }

    def test_bad_model(self, webserver):
        with pytest.raises(RegistryManagerException):
            webserver.get('/bad/model')


class TestBlok:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        return

    def test_predicate(self, webserver):
        webserver.get('/bloks1', status=404)
        webserver.get('/bloks2', status=200)

    def test_saved_errors_in_request(self, registry_testblok):
        registry = registry_testblok

        class Error:
            def __init__(self):
                self._errors = []
                self.status = 0

            def add(self, *a):
                self._errors.append(a)

        class A:
            def __init__(self):
                self.registry = registry

        class Request:
            anyblok = A()
            errors = Error()

        request = Request()
        with saved_errors_in_request(request):
            raise Exception()

        assert request.errors.status == 500


class TestCrudResourceBlokModelSchemaValidator:
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """
    collection_path = '/bloks/v5'
    path = '/bloks/v5/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_5):
        transaction = registry_rest_api_5.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_model_schema_validator_get(self, registry_rest_api_5, webserver):
        response = webserver.get(self.path % 'anyblok-core')
        assert response.status_code == 200
        assert response.json_body == {
            'author': 'Suzanne Jean-Sébastien',
            'name': 'anyblok-core',
            'order': 0,
            'state': 'installed',
        }



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice.resource import resource
from anyblok_pyramid import current_blok

from anyblok_pyramid_rest_api.crud_resource import (
    CrudResource
)

from .schema import (
    CustomerSchema,
    AddressSchema,
)


@resource(
    collection_path='/customers/v5',
    path='/customers/v5/{id}',
    installed_blok=current_blok()
)
class CustomerResourceV5(CrudResource):
    model = 'Model.Customer'
    default_schema = CustomerSchema


@resource(
    collection_path='/addresses/v5',
    path='/addresses/v5/{id}',
    installed_blok=current_blok()
)
class AddressResourceV5(CrudResource):
    model = 'Model.Address'
    default_schema = AddressSchema


@resource(
    collection_path='/bloks/v5',
    path='/bloks/v5/{name}',
    installed_blok=current_blok()
)
class BlokResourceV5(CrudResource):
    model = 'Model.System.Blok'

    @classmethod
    def get_serialize_opts(cls, rest_action):
        opts = super(BlokResourceV5, cls).get_serialize_opts(rest_action)
        opts['only'] = ['author', 'order', 'name', 'state']
        return opts



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from marshmallow import Schema, fields
from anyblok_pyramid_rest_api.schema import FullRequestSchema


class ExampleSchema(Schema):
    """A basic marshmallow schema example
    This one represents the model.Example fields
    """
    id = fields.Int(required=True)
    name = fields.Str(required=True)


class ExamplePathSchema(Schema):
    """A basic marshmallow schema example
    This one represents the model.Example fields
    """
    id = fields.Int(required=True)


class ThingSchema(Schema):
    """Schema for the Thing model
    """
    uuid = fields.UUID(required=True)
    name = fields.Str(required=True)
    secret = fields.Str(required=True, load_only=True)
    example_id = fields.Int(required=True)
    example = fields.Nested(ExampleSchema)
    create_date = fields.DateTime(dump_only=True)
    update_date = fields.DateTime(dump_only=True)


class ThingRequestSchema(FullRequestSchema):
    """This one inherits FullRequestSchema and represents the request
    model.Thing
    """
    body = fields.Nested(ThingSchema(only=(
        'name', 'secret', 'example_id', 'example', 'create_date',
        'update_date')))
    path = fields.Nested(ThingSchema(only=('uuid',)))


class AnotherSchema(FullRequestSchema):
    """This one inherits FullRequestSchema and represents the request
    model.Example
    """
    body = fields.Nested(ExampleSchema, only=('name',))
    path = fields.Nested(ExampleSchema, only=('id',))



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok_pyramid_rest_api.querystring import QueryString
from anyblok.column import Integer, String
from anyblok.relationship import Many2One
from .conftest import init_registry_with_bloks


def add_integer_class():

    from anyblok import Declarations

    @Declarations.register(Declarations.Model)
    class Exemple:
        id = Integer(primary_key=True)
        number = Integer()


def add_many2one_class():

    from anyblok import Declarations

    @Declarations.register(Declarations.Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @Declarations.register(Declarations.Model)
    class Test2:
        id = Integer(primary_key=True)
        test = Many2One(model=Declarations.Model.Test)


class MockRequestError:

    def __init__(self, testcase):
        self.testcase = testcase
        self.messages = []

    def add(self, a1, a2, a3):
        assert a1 == 'querystring'
        assert a2 == '400 Bad Request'
        self.messages.append(a3)


class MockRequest:

    def __init__(self, testcase):
        self.errors = MockRequestError(testcase)
        self.params = {}


class TestQueryString:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok):
        transaction = registry_blok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_querystring_update_filter_eq(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'eq'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        obj = Q.one()
        assert obj.name == 'anyblok-core'

    def test_querystring_update_sqlalchemy_query(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        qs.update_sqlalchemy_query(query)

    def test_querystring_update_filter_like(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'like'
        value = 'yblok-c'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        obj = Q.one()
        assert obj.name == 'anyblok-core'

    def test_querystring_update_filter_ilike(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'ilike'
        value = 'yBlOk-c'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        obj = Q.one()
        assert obj.name == 'anyblok-core'

    def test_querystring_update_filter_or_ilike(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'or-ilike'
        value = 'core,test'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        Q = Q.order_by('name')
        obj = Q.all()
        assert obj.name == ['anyblok-core', 'anyblok-test']

    def test_querystring_update_filter_or_without_value(self,
                                                        registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        key = 'name'
        op = 'or-ilike'
        value = None
        qs = QueryString(request, model)
        assert qs.update_or_filter(model, key, op, value) is None

    def test_querystring_update_filter_in_1(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'in'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        obj = Q.all()
        assert obj.name == ['anyblok-core']

    def test_querystring_update_filter_in_2(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'in'
        value = 'anyblok-core,anyblok-test'
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        names = Q.all().name
        assert len(names) == 2
        assert 'anyblok-core' in names
        assert 'anyblok-test' in names

    def test_querystring_update_filter_in_3(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        key = 'name'
        op = 'in'
        value = ''
        qs = QueryString(request, model)
        qs.update_filter(model, key, op, value)
        assert (
            "Filter 'in' except a comma separated string value" in
            request.errors.messages)

    def test_querystring_from_filter_by_ok(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'eq'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        Q = qs.from_filter_by(query)
        obj = Q.one()
        assert obj.name == 'anyblok-core'

    def test_querystring_from_filter_by_without_key(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = None
        op = 'eq'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        qs.from_filter_by(query)
        assert len(request.errors.messages) == 1

    def test_querystring_from_filter_by_ko_bad_op(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'unknown'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        qs.from_filter_by(query)
        assert (
            "Filter 'unknown' does not exist." in
            request.errors.messages)

    def test_querystring_from_filter_by_bad_key(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'badkey'
        op = 'eq'
        value = 'anyblok-core'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        qs.from_filter_by(query)
        assert (
            "Filter 'badkey': 'badkey' does not exist in model "
            "<class 'anyblok.model.factory.ModelSystemBlok'>." in
            request.errors.messages)

    def test_querystring_from_order_by_ok(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'asc'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        Q = qs.from_order_by(query)
        assert Q.all().name == query.order_by(model.name.asc()).all().name

    def test_querystring_from_order_by_without_key(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = None
        op = 'asc'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        qs.from_order_by(query)
        assert len(request.errors.messages) == 1

    def test_querystring_from_order_by_ko_bad_op(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'name'
        op = 'unknown'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        qs.from_order_by(query)
        assert (
            "ORDER_by operator 'unknown' does not exist." in
            request.errors.messages)

    def test_querystring_from_order_by_bad_key(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        key = 'badkey'
        op = 'asc'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        qs.from_order_by(query)
        assert (
            "Order 'badkey': 'badkey' does not exist in model "
            "<class 'anyblok.model.factory.ModelSystemBlok'>." in
            request.errors.messages)

    def test_querystring_from_limit(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        qs.limit = 1
        Q = qs.from_limit(query)
        assert len(Q.all()) == 1

    def test_querystring_from_limit_without_limit(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        qs.limit = None
        Q = qs.from_limit(query)
        assert len(Q.all()) == len(query.all())

    def test_querystring_from_offset(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        qs.offset = 3
        Q = qs.from_offset(query)
        assert len(Q.all()) == len(query.offset(3).all())

    def test_querystring_from_offset_without_offset(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        qs.offset = None
        Q = qs.from_offset(query)
        assert len(Q.all()) == len(query.all())

    def test_querystring_get_model_and_key_from_relationship_1(self,
                                                               registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        query = model.query()
        qs = QueryString(request, model)
        res = qs.get_model_and_key_from_relationship(query, model, ['name'])
        assert res[1] is model
        assert res[2] == 'name'

    def test_querystring_get_model_and_key_from_relationship_4(self,
                                                               registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Column
        query = model.query()
        qs = QueryString(request, model)
        res = qs.get_model_and_key_from_relationship(
            query, model, ['model', 'name'])
        assert (
            "'model' in model "
            "<class 'anyblok.model.factory.ModelSystemColumn'> is not "
            "a relationship." == res)

    def test_has_no_tag(self, registry_blok):
        registry = registry_blok
        request = MockRequest(self)
        model = registry.System.Blok
        qs = QueryString(request, model)
        assert not (qs.has_tag('unknown.tag'))


@pytest.fixture(scope="class")
def registry_blok_with_integer(request, bloks_loaded):
    registry = init_registry_with_bloks([], add_integer_class)
    request.addfinalizer(registry.close)
    return registry


class TestQueryStringOnInteger:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok_with_integer):
        transaction = registry_blok_with_integer.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_querystring_update_filter_lt(self, registry_blok_with_integer):
        registry = registry_blok_with_integer
        request = MockRequest(self)
        model = registry.Exemple
        query = model.query()
        key = 'number'
        op = 'lt'
        value = 10
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        assert len(Q.all()) == 0
        model.insert(number=9)
        assert len(Q.all()) == 1
        model.insert(number=10)
        assert len(Q.all()) == 1
        model.insert(number=11)
        assert len(Q.all()) == 1

    def test_querystring_update_filter_lte(self, registry_blok_with_integer):
        registry = registry_blok_with_integer
        request = MockRequest(self)
        model = registry.Exemple
        query = model.query()
        key = 'number'
        op = 'lte'
        value = 10
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        assert len(Q.all()) == 0
        model.insert(number=9)
        assert len(Q.all()) == 1
        model.insert(number=10)
        assert len(Q.all()) == 2
        model.insert(number=11)
        assert len(Q.all()) == 2

    def test_querystring_update_filter_gt(self, registry_blok_with_integer):
        registry = registry_blok_with_integer
        request = MockRequest(self)
        model = registry.Exemple
        query = model.query()
        key = 'number'
        op = 'gt'
        value = 10
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        assert len(Q.all()) == 0
        model.insert(number=9)
        assert len(Q.all()) == 0
        model.insert(number=10)
        assert len(Q.all()) == 0
        model.insert(number=11)
        assert len(Q.all()) == 1

    def test_querystring_update_filter_gte(self, registry_blok_with_integer):
        registry = registry_blok_with_integer
        request = MockRequest(self)
        model = registry.Exemple
        query = model.query()
        key = 'number'
        op = 'gte'
        value = 10
        qs = QueryString(request, model)
        Q = query.filter(qs.update_filter(model, key, op, value))
        assert len(Q.all()) == 0
        model.insert(number=9)
        assert len(Q.all()) == 0
        model.insert(number=10)
        assert len(Q.all()) == 1
        model.insert(number=11)
        assert len(Q.all()) == 2


@pytest.fixture(scope="class")
def registry_blok_with_m2o(request, bloks_loaded):
    registry = init_registry_with_bloks([], add_many2one_class)
    request.addfinalizer(registry.close)
    return registry


class TestQueryStringWithM2O:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok_with_m2o):
        transaction = registry_blok_with_m2o.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_querystring_from_filter_by_with_relationship(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        key = 'test.name'
        op = 'eq'
        value = 'test'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        Q = qs.from_filter_by(query)
        assert len(Q.all()) == 0
        t1 = registry.Test(name='test')
        t2 = registry.Test(name='other')
        model.insert(test=t1)
        assert len(Q.all()) == 1
        model.insert(test=t2)
        assert len(Q.all()) == 1
        model.insert(test=t2)
        assert len(Q.all()) == 1
        model.insert(test=t1)
        assert len(Q.all()) == 2

    def test_querystring_from_filter_by_with_relationship_bad_key(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        key = 'test.badkey'
        op = 'eq'
        value = 'test'
        qs = QueryString(request, model)
        qs.filter_by = [dict(key=key, op=op, value=value)]
        qs.from_filter_by(query)
        assert (
            "Filter 'test.badkey': 'badkey' does not exist in model "
            "<class 'anyblok.model.factory.ModelTest'>." in
            request.errors.messages)

    def test_querystring_from_order_by_ok_with_relationship(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        key = 'test.name'
        op = 'asc'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        Q = qs.from_order_by(query)
        t1 = registry.Test(name='test')
        t2 = registry.Test(name='other')
        model.insert(test=t1)
        model.insert(test=t2)
        model.insert(test=t2)
        model.insert(test=t1)
        query = query.join(registry.Test2.test)
        query = query.order_by(registry.Test.name.asc())
        assert Q.all() == query.all()

    def test_querystring_from_order_by_bad_key_relationship(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        key = 'test.badkey'
        op = 'asc'
        qs = QueryString(request, model)
        qs.order_by = [dict(key=key, op=op)]
        qs.from_order_by(query)
        assert (
            "Order 'test.badkey': 'badkey' does not exist in model "
            "<class 'anyblok.model.factory.ModelTest'>." in
            request.errors.messages)

    def test_querystring_get_remote_model_for(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.System.Blok
        qs = QueryString(request, model)
        Model = qs.get_remote_model_for(registry.Test2, 'test')
        assert Model is registry.Test

    def test_querystring_get_remote_model_for_without_relationship(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.System.Blok
        qs = QueryString(request, model)
        Model = qs.get_remote_model_for(registry.Test, 'name')
        assert Model is None

    def test_querystring_get_remote_model_for_unknown(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.System.Blok
        qs = QueryString(request, model)
        Model = qs.get_remote_model_for(registry.Test, 'test')
        assert Model is None

    def test_querystring_get_model_and_key_from_relationship_2(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        qs = QueryString(request, model)
        res = qs.get_model_and_key_from_relationship(
            query, model, ['test', 'name'])
        assert res[1] is registry.Test
        assert res[2] == 'name'

    def test_querystring_get_model_and_key_from_relationship_3(
        self, registry_blok_with_m2o
    ):
        registry = registry_blok_with_m2o
        request = MockRequest(self)
        model = registry.Test2
        query = model.query()
        qs = QueryString(request, model)
        res = qs.get_model_and_key_from_relationship(
            query, model, ['test', 'number'])
        waiting = ("'number' does not exist in model <class 'anyblok.model."
                   "factory.ModelTest'>.")
        assert res == waiting



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import logging
from copy import deepcopy
import pytest
from anyblok.blok import BlokManager
from anyblok.config import Configuration
from anyblok.environment import EnvironmentManager
from anyblok.registry import RegistryManager
from copy import copy
import sqlalchemy
from sqlalchemy_utils.functions import database_exists, create_database, orm
import os

logger = logging.getLogger(__name__)


def init_registry_with_bloks(bloks, function, **kwargs):
    if bloks is None:
        bloks = []
    if isinstance(bloks, tuple):
        bloks = list(bloks)
    if isinstance(bloks, str):
        bloks = [bloks]

    anyblok_test_name = 'anyblok-test'
    if anyblok_test_name not in bloks:
        bloks.append(anyblok_test_name)

    loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
    if function is not None:
        EnvironmentManager.set('current_blok', anyblok_test_name)
        try:
            function(**kwargs)
        finally:
            EnvironmentManager.set('current_blok', None)
    try:
        registry = RegistryManager.get(
            Configuration.get('db_name'),
            unittest=True)

        # update required blok
        registry_bloks = registry.get_bloks_by_states('installed', 'toinstall')
        toinstall = [x for x in bloks if x not in registry_bloks]
        toupdate = [x for x in bloks if x in registry_bloks]
        registry.upgrade(install=toinstall, update=toupdate)
    finally:
        RegistryManager.loaded_bloks = loaded_bloks

    return registry


def init_registry(function, **kwargs):
    return init_registry_with_bloks([], function, **kwargs)


def drop_database(url):
    url = copy(sqlalchemy.engine.url.make_url(url))
    database = url.database
    if url.drivername.startswith('postgresql'):
        url.database = 'postgres'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    engine = sqlalchemy.create_engine(url)
    if engine.dialect.name == 'sqlite' and url.database != ':memory:':
        os.remove(url.database)
    else:
        text = 'DROP DATABASE {0}'.format(orm.quote(engine, database))
        cnx = engine.connect()
        cnx.execute("ROLLBACK")
        cnx.execute(text)
        cnx.execute("commit")
        cnx.close()


@pytest.fixture(scope="session")
def base_loaded(request):
    url = Configuration.get('get_url')()
    if not database_exists(url):
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    BlokManager.load()
    registry = init_registry_with_bloks([], None)
    registry.commit()
    registry.close()
    BlokManager.unload()


@pytest.fixture(scope="module")
def bloks_loaded(request, base_loaded):
    request.addfinalizer(BlokManager.unload)
    BlokManager.load()


@pytest.fixture(scope="module")
def testbloks_loaded(request, base_loaded):
    request.addfinalizer(BlokManager.unload)
    BlokManager.load(entry_points=('bloks', 'test_bloks'))


@pytest.fixture(scope="class")
def registry_blok(request, bloks_loaded):
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_testblok(request, testbloks_loaded):
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="function")
def registry_testblok_func(request, testbloks_loaded):
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry


# Pyramid
from webtest import TestApp  # noqa
from anyblok_pyramid.pyramid_config import Configurator  # noqa


@pytest.fixture(scope="class")
def webserver(request):
    config = Configurator()
    config.include_from_entry_point()
    # No param here # for includeme in self.includemes:
    # No param here #     config.include(includeme)

    config.load_config_bloks()
    app = config.make_wsgi_app()
    return TestApp(app)


# rest api


@pytest.fixture(scope="class")
def registry_rest_api_1(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_1'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_2(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_2'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_3(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_3'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_4(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_4'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_5(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_5'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_5_logged(request, testbloks_loaded):
    registry = init_registry_with_bloks(
        ['test_rest_api_5', 'auth-password'], None)
    request.addfinalizer(registry.close)

    registry.User.insert(
        login='jssuzanne', first_name='js', last_name='Suzanne')
    registry.User.CredentialStore.insert(
        login='jssuzanne', password='mypassword')
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_6(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_6'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_7(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_7'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_8(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_8'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_9(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_9'], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_rest_api_10(request, testbloks_loaded):
    registry = init_registry_with_bloks(['test_rest_api_10'], None)
    request.addfinalizer(registry.close)
    return registry



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""A set of reusable basic schemas"""
from marshmallow import Schema, fields


class FullRequestSchema(Schema):
    """A base marshmallow schema to validate the entire request
    :Example:
        >>> from marshmallow import Schema, fields
        >>> from anyblok_pyramid_rest_api.schema import FullRequestSchema
        >>> class ModelSchema(Schema):
        >>>     id = fields.Int(required=True)
        >>>     name = fields.Str(required=True)
        >>> class RequestModelSchema(FullRequestSchema):
        >>>     body = fields.Nested(ModelSchema(partial=('id',)))
        >>>     path = fields.Nested(ModelSchema(partial=('name',)))
    """
    body = fields.Dict()
    url = fields.Str()
    path = fields.Dict()
    method = fields.Str()
    querystring = fields.Dict()
    headers = fields.Dict()
    cookies = fields.Dict()



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sébastien Suzanne <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_marshmallow import SchemaWrapper
from marshmallow.schema import Schema
from anyblok_marshmallow.fields import Nested, String


class CitySchema(SchemaWrapper):
    model = 'Model.City'


class TagSchema(SchemaWrapper):
    model = 'Model.Tag'


class AddressSchema(SchemaWrapper):
    model = 'Model.Address'

    class Schema:
        # follow the relationship Many2One and One2One
        city = Nested(CitySchema)


class CustomerSchema(SchemaWrapper):
    """Schema for 'Model.Customer'
    """
    model = 'Model.Customer'

    class Schema:
        # follow the relationship One2Many and Many2Many
        # - the many=True is required because it is *2Many
        # - exclude is used to forbid the recurse loop
        addresses = Nested(AddressSchema, many=True, exclude=('customer', ))
        tags = Nested(TagSchema, many=True)


class ActionSchema(Schema):
    name = String(required=True)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2018 Jean-Sébastien Suzanne <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid import current_blok
from anyblok_pyramid_rest_api.crud_resource import (
    CrudResource, resource)
from .schema import CustomerSchema, ActionSchema


@resource(
    collection_path='/customers/v8',
    path='/customers/v8/{id}',
    installed_blok=current_blok()
)
class CustomerResourceV8(CrudResource):
    model = 'Model.Customer'
    default_schema = CustomerSchema

    @CrudResource.service('action1', verb='GET', collection=True)
    def do_action_1(self):
        return 'test'

    @CrudResource.service('action2', schema=ActionSchema)
    def do_action_2(self):
        return self.body['name']



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .validator import (
    FILTER_OPERATORS, ORDER_BY_OPERATORS, deserialize_querystring
)
from sqlalchemy import or_
from logging import getLogger
logger = getLogger(__name__)


class QueryString:

    def __init__(self, request, Model, adapter=None):
        self.request = request
        self.adapter = adapter
        self.Model = Model
        if request.params is not None:
            parsed_params = deserialize_querystring(request.params)
            self.filter_by = parsed_params.get('filter_by', [])
            self.tags = parsed_params.get('tags')
            self.order_by = parsed_params.get('order_by', [])
            self.context = parsed_params.get('context', {})
            self.limit = parsed_params.get('limit')
            if self.limit and isinstance(self.limit, str):
                self.limit = int(self.limit)

            self.offset = parsed_params.get('offset')
            if self.offset and isinstance(self.offset, str):
                self.offset = int(self.offset)

    def update_sqlalchemy_query(self, query, only_filter=False):
        query = self.from_filter_by(query)
        query = self.from_tags(query)
        if not only_filter:
            query = self.from_order_by(query)
            query = self.from_limit(query)
            query = self.from_offset(query)

        return query

    def from_filter_by(self, query):
        for item in self.filter_by:
            op = item.get('op')
            key = item.get('key')
            value = item.get('value')
            mode = item.get('mode', 'include')
            # Is operator valid?
            if op not in FILTER_OPERATORS:
                self.request.errors.add(
                    'querystring',
                    '400 Bad Request', 'Filter %r does not exist.' % op)
                self.request.errors.status = 400
            elif not key:
                self.request.errors.add(
                    'querystring',
                    '400 Bad Request',
                    "No key filled %r" % item)
                self.request.errors.status = 400
            elif self.has_specific_filter(key, op):
                query = self.specific_filter(query, key, op, value, mode)
            else:
                res = self.get_model_and_key_from_relationship(
                    query, self.Model, key.split('.'))
                if isinstance(res, tuple):
                    _query, _model, _key = res
                    condition = self.update_filter(_model, _key, op, value)
                    if condition is not None:
                        if mode == 'include':
                            query = _query.filter(condition)
                        elif mode == 'exclude':
                            query = _query.filter(~condition)

                    if '.' in key:
                        query = query.reset_joinpoint()
                else:
                    self.request.errors.add(
                        'querystring',
                        '400 Bad Request',
                        "Filter %r: %s" % (key, res))
                    self.request.errors.status = 400

        return query

    def has_specific_filter(self, key, op):
        if self.adapter is None:
            return False

        return self.adapter.has_filter_for(key, op)

    def specific_filter(self, query, key, op, value, mode):
        try:
            return self.adapter.get_filter_for(key, op)(
                self, query, op, value, mode)
        except Exception as e:
            self.request.errors.add(
                'querystring',
                "Filter %s %s %r" % (key, op, value),
                str(e)
            )
            self.request.errors.status = 400
            logger.exception(str(e))

        return query

    def has_specific_order_by(self, key):
        if self.adapter is None:
            return False

        return self.adapter.has_order_by_for(key)

    def specific_order_by(self, query, key, op):
        try:
            return self.adapter.get_order_by_for(key)(self, query, op)
        except Exception as e:
            self.request.errors.add(
                'querystring',
                "Order by %s %s" % (key, op),
                str(e)
            )
            self.request.errors.status = 400
            logger.exception(str(e))

        return query

    def from_tags(self, query):
        for tag in self.tags:
            if self.has_tag(tag):
                try:
                    query = self.from_tag(query, tag)
                except Exception as e:
                    self.request.errors.add(
                        'querystring',
                        "Tag %r" % tag,
                        str(e)
                    )
                    self.request.errors.status = 400
                    logger.exception(str(e))
            else:
                self.request.errors.add(
                    'querystring',
                    "Tag %r" % tag,
                    "Unexisting tag"
                )
                self.request.errors.status = 400

        return query

    def has_tag(self, tag):
        if self.adapter is None:
            return False

        return self.adapter.has_tag_for(tag)

    def from_tag(self, query, tag):
        return self.adapter.get_tag_for(tag)(self, query)

    def from_order_by(self, query):
        for item in self.order_by:
            op = item.get('op')
            key = item.get('key')
            # Is operator valid?
            if op not in ORDER_BY_OPERATORS:
                self.request.errors.add(
                    'querystring', '400 Bad Request',
                    "ORDER_by operator '%s' does not exist." % op)
                self.request.errors.status = 400
            # Column exists?
            elif not key:
                self.request.errors.add(
                    'querystring',
                    '400 Bad Request',
                    "No key ordered %r" % item)
                self.request.errors.status = 400
            elif self.has_specific_order_by(key):
                query = self.specific_order_by(query, key, op)
            else:
                res = self.get_model_and_key_from_relationship(
                    query, self.Model, key.split('.'))
                if isinstance(res, tuple):
                    _query, _model, _key = res
                    query = _query.order_by(
                        getattr(getattr(_model, _key), op)())

                    if '.' in key:
                        query = query.reset_joinpoint()
                else:
                    self.request.errors.add(
                        'querystring',
                        '400 Bad Request',
                        "Order %r: %s" % (key, res))
                    self.request.errors.status = 400

        return query

    def from_limit(self, query):
        if self.limit:
            query = query.limit(self.limit)

        return query

    def from_offset(self, query):
        if self.offset:
            query = query.offset(self.offset)

        return query

    def update_or_filter(self, model, key, op, value):
        if not value:
            self.request.errors.add(
                'querystring', '400 Bad Request',
                'not splitting entries for %r: %r' % (key, value)
            )
            self.request.errors.status = 400
            return
        return or_(*[
            self.update_filter(model, key, op, v.strip())
            for v in value.split(',')
        ])

    def update_filter(self, model, key, op, value):
        if op == "eq":
            return getattr(model, key) == value
        elif op in ("like", "ilike"):
            return getattr(getattr(model, key), op)("%" + value + "%")
        elif op == "lt":
            return getattr(model, key) < value
        elif op == "lte":
            return getattr(model, key) <= value
        elif op == "gt":
            return getattr(model, key) > value
        elif op == "gte":
            return getattr(model, key) >= value
        elif op == "in":
            # ensure we have a comma separated value string...
            if value:
                values = value.split(',')
                return getattr(model, key).in_(values)
            error = 'Filter %r except a comma separated string value' % op
        elif op.startswith("or-"):
            return self.update_or_filter(model, key, op.split('-')[1], value)

        self.request.errors.add('querystring', '400 Bad Request', error)
        self.request.errors.status = 400

    def get_remote_model_for(self, Model, fieldname):
        registry = Model.registry
        Field = registry.System.Field
        query = Field.query()
        models = [Model.__registry_name__]
        for base in Model.__anyblok_bases__:
            models.append(base.__registry_name__)

        query = query.filter(Field.model.in_(models))
        query = query.filter(Field.name == fieldname)
        field = query.first()
        if field and field.remote_model:
            return registry.get(field.remote_model)

        return None

    def get_model_and_key_from_relationship(self, query, model, keys,
                                            already_join=False):
        key = keys[0]
        if not hasattr(model, 'fields_description'):
            return '%r is not an SQL Model you should use Adapter' % model
        if key not in model.fields_description():
            return '%r does not exist in model %s.' % (key, model)

        if len(keys) == 1:
            return (query, model, keys[0])
        else:
            field = getattr(model, key)
            new_model = self.get_remote_model_for(model, key)
            if new_model is None:
                return '%r in model %s is not a relationship.' % (key, model)

            query = query.join(field, aliased=True, from_joinpoint=already_join)
            return self.get_model_and_key_from_relationship(
                query, new_model, keys[1:], already_join=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger

logger = getLogger(__name__)


class Adapter:

    def __init__(self, registry, Model):
        self.registry = registry
        self.Model = Model
        self.loaded = False
        self.filters = {}
        self.orders_by = {}
        self.tags = {}

    def load_decorators(self):
        for attr, value in self.__class__.__dict__.items():
            if hasattr(value, 'is_filter'):
                key, operators = value.is_filter
                self.filters[key] = {operator: attr for operator in operators}
            elif hasattr(value, 'is_order_by'):
                self.orders_by[value.is_order_by] = attr
            elif hasattr(value, 'is_tag'):
                self.tags[value.is_tag] = attr

    def has_filter_for(self, key, operator):
        return True if self.filters.get(key, {}).get(operator) else False

    def get_filter_for(self, key, operator):
        return getattr(self, self.filters[key][operator])

    def has_order_by_for(self, key):
        return True if self.orders_by.get(key) else False

    def get_order_by_for(self, key):
        return getattr(self, self.orders_by[key])

    def has_tag_for(self, tag):
        return True if self.tags.get(tag) else False

    def get_tag_for(self, tag):
        return getattr(self, self.tags[tag])

    @classmethod
    def filter(cls, key, operators):
        if not isinstance(operators, (list, tuple)):
            operators = [operators]

        def wrapper(method):
            method.is_filter = (key, operators)
            return method

        return wrapper

    @classmethod
    def order_by(cls, name):
        def wrapper(method):
            method.is_order_by = name
            return method

        return wrapper

    @classmethod
    def tag(cls, name):
        def wrapper(method):
            method.is_tag = name
            return method

        return wrapper



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice.resource import resource
from anyblok_pyramid import current_blok
from anyblok_pyramid_rest_api.crud_resource import CrudResource
from anyblok_pyramid_rest_api.adapter import Adapter as Adapter
from .schema import CustomerSchema
from sqlalchemy import or_


@resource(
    collection_path='/customers/v6',
    path='/customers/v6/{id}',
    installed_blok=current_blok()
)
class CustomerResourceV6(CrudResource):
    model = 'Model.Customer'
    default_schema = CustomerSchema

    class adapter_cls(Adapter):

        @Adapter.filter('addresses.city', ['ilike'])
        def filter_by_location_and_children(self, querystring, query, operator,
                                            value, mode):
            query = query.join(self.registry.Customer.addresses, aliased=True)
            query = query.join(self.registry.Address.city, aliased=True)
            query = query.filter(or_(
                self.registry.City.name.ilike(value),
                self.registry.City.zipcode.ilike(value)))
            return query

        @Adapter.tag('green')
        def tag_is_green(self, querystring, query):
            query = query.join(self.registry.Customer.tags, aliased=True)
            query = query.filter(self.registry.Tag.name == 'green')
            return query

        @Adapter.tag('orange')
        def tag_is_orange(self, querystring, query):
            query = query.join(self.registry.Customer.tags, aliased=True)
            query = query.filter(self.registry.Tag.name == 'orange')
            return query

        @Adapter.tag('colors')
        def tag_color_depend_of_context(self, querystring, query):
            colors = querystring.context.get('colors')
            if colors:
                colors = colors.split(',')
                query = query.join(self.registry.Customer.tags, aliased=True)
                query = query.filter(self.registry.Tag.name.in_(colors))

            return query

        @Adapter.tag('wrong')
        def wrong_tag(self, querystring, query):
            raise Exception('Wrong tags')

        @Adapter.filter('addresses.other', ['ilike'])
        def wrong_filter(self, querystring, query, operator,
                         value, mode):
            raise Exception('Wrong filters')

        @Adapter.order_by('address')
        def order_by_city(self, querystring, query, operator):
            query = query.join(self.registry.Customer.addresses, aliased=True)
            query = query.join(self.registry.Address.city, aliased=True)
            return query.order_by(self.registry.City.zipcode)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice.resource import resource
from anyblok_pyramid import current_blok
from anyblok_pyramid_rest_api.crud_resource import CrudResource
from anyblok_pyramid_rest_api.adapter import Adapter
from .schema import CustomerSchema
from sqlalchemy import or_


class CustomerAdapter(Adapter):

    @Adapter.filter('addresses.city', ['ilike'])
    def filter_by_location_and_children(self, querystring, query, operator,
                                        value, mode):
        query = query.join(self.registry.Customer.addresses, aliased=True)
        query = query.join(self.registry.Address.city, aliased=True)
        query = query.filter(or_(
            self.registry.City.name.ilike(value),
            self.registry.City.zipcode.ilike(value)))
        return query

    @Adapter.tag('green')
    def tag_is_green(self, querystring, query):
        query = query.join(self.registry.Customer.tags, aliased=True)
        query = query.filter(self.registry.Tag.name == 'green')
        return query

    @Adapter.tag('orange')
    def tag_is_orange(self, querystring, query):
        query = query.join(self.registry.Customer.tags, aliased=True)
        query = query.filter(self.registry.Tag.name == 'orange')
        return query

    @Adapter.tag('colors')
    def tag_color_depend_of_context(self, querystring, query):
        colors = querystring.context.get('colors')
        if colors:
            colors = colors.split(',')
            query = query.join(self.registry.Customer.tags, aliased=True)
            query = query.filter(self.registry.Tag.name.in_(colors))

        return query

    @Adapter.tag('wrong')
    def wrong_tag(self, querystring, query):
        raise Exception('Wrong tags')

    @Adapter.filter('addresses.other', ['ilike'])
    def wrong_filter(self, querystring, query, operator,
                     value, mode):
        raise Exception('Wrong filters')

    @Adapter.order_by('wrong')
    def order_by_wrong(self, querystring, query, operator):
        raise Exception('Wrong tags')

    @Adapter.order_by('address')
    def order_by_city(self, querystring, query, operator):
        query = query.join(self.registry.Customer.addresses, aliased=True)
        query = query.join(self.registry.Address.city, aliased=True)
        return query.order_by(self.registry.City.zipcode)


@resource(
    collection_path='/customers/v4',
    path='/customers/v4/{id}',
    installed_blok=current_blok()
)
class CustomerResourceV4(CrudResource):
    model = 'Model.Customer'
    adapter_cls = CustomerAdapter
    default_schema = CustomerSchema



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice.resource import resource
from anyblok_pyramid import current_blok
from anyblok_pyramid_rest_api.crud_resource import CrudResource
from .schema import CustomerSchema, AddressSchema


@resource(
    collection_path='/customers/v3',
    path='/customers/v3/{id}',
    installed_blok=current_blok()
)
class CustomerResourceV3(CrudResource):
    model = 'Model.Customer'
    default_schema = CustomerSchema


@resource(
    collection_path='/addresses/v3',
    path='/addresses/v3/{id}',
    installed_blok=current_blok()
)
class AddressResourceV3(CrudResource):
    model = 'Model.Address'
    default_schema = AddressSchema



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid import current_blok
from cornice.resource import resource
from anyblok_pyramid_rest_api.crud_resource import CrudResource


@resource(
    collection_path='/bloks1',
    path='/blok1/{name}',
    installed_blok=current_blok()
)
class CrudBlok(CrudResource):
    model = 'Model.System.Blok'


@resource(
    collection_path='/bloks2',
    path='/blok2/{name}',
)
class CrudBlok2(CrudResource):
    model = 'Model.System.Blok'


@resource(
    collection_path='/columns',
    path='/column/{model}/{name}',
    installed_blok=current_blok()
)
class CrudColumn(CrudResource):
    model = 'Model.System.Column'


@resource(
    collection_path='/bad/model',
    path='/bad/model/{id}',
)
class CrudBadModel(CrudResource):
    model = 'Model.Bad.Model'



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer, String
from anyblok.relationship import Many2One, Many2Many


Model = Declarations.Model


@Declarations.register(Declarations.Model)
class City:

    id = Integer(primary_key=True)
    name = String(nullable=False)
    zipcode = String(nullable=False)

    def __repr__(self):
        return '<City(name={self.name!r})>'.format(self=self)


@Declarations.register(Declarations.Model)
class Tag:

    id = Integer(primary_key=True)
    name = String(nullable=False)

    def __repr__(self):
        return '<Tag(name={self.name!r})>'.format(self=self)


@Declarations.register(Declarations.Model)
class Customer:
    id = Integer(primary_key=True)
    name = String(nullable=False)
    tags = Many2Many(model=Declarations.Model.Tag)

    def __repr__(self):
        return '<Customer(name={self.name!r}, tags={self.tags!r})>'.format(
            self=self)


@Declarations.register(Declarations.Model)
class Address:

    id = Integer(primary_key=True)
    street = String(nullable=False)
    city = Many2One(model=Declarations.Model.City, nullable=False)
    customer = Many2One(
        model=Declarations.Model.Customer, nullable=False,
        foreign_key_options={'ondelete': 'cascade'}, one2many="addresses")



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice.validators import extract_cstruct
from marshmallow import ValidationError, INCLUDE
from logging import getLogger
import re

logger = getLogger(__name__)


FILTER_OPERATORS = [
    'eq', 'like', 'ilike', 'lt', 'lte', 'gt', 'gte',
    'in', 'or-like', 'or-ilike', 'or-lt', 'or-lte', 'or-gt', 'or-gte'
]
ORDER_BY_OPERATORS = ['asc', 'desc']


def parse_key_with_two_elements(filter_):
    pattern = ".*\\[(.*)\\]\\[(.*)\\]"
    return re.match(pattern, filter_).groups()


def parse_key_with_one_element(filter_):
    pattern = ".*\\[(.*)\\]"
    return re.match(pattern, filter_).groups()[0]


def deserialize_querystring(params=None):
    """
    Given a querystring parameters dict, returns a new dict that will be used
    to build query filters.
    The logic is to keep everything but transform some key, values to build
    database queries.
    Item whose key starts with 'filter[*' will be parsed to a key, operator,
    value dict (filter_by).
    Item whose key starts with 'order_by[*' will be parse to a key, operator
    dict(order_by).
    'limit' and 'offset' are kept as is.
    All other keys are added to 'filter_by' with 'eq' as default operator.

    # TODO: Use marshmallow pre-validation feature
    # TODO: Evaluate 'webargs' python module to see if it can helps

    :param params: A dict that represent a querystring (request.params)
    :type params: dict
    :return: A suitable dict for building a filtering query
    :rtype: dict
    """
    filter_by = []
    order_by = []
    tags = []
    context = {}
    limit = None
    offset = 0
    for param in params.items():
        k, v = param
        # TODO  better regex or something?
        if k.startswith("filter["):
            # Filtering (include)
            key, op = parse_key_with_two_elements(k)
            filter_by.append(dict(key=key, op=op, value=v, mode="include"))
        elif k.startswith("~filter["):
            # Filtering (exclude)
            # TODO check for errors into string pattern
            key, op = parse_key_with_two_elements(k)
            filter_by.append(dict(key=key, op=op, value=v, mode="exclude"))
        elif k.startswith("context["):
            key = parse_key_with_one_element(k)
            context[key] = v
        elif k == "tag":
            tags.append(v)
        elif k == "tags":
            tags.extend(v.split(','))
        elif k.startswith("order_by["):
            # Ordering
            op = parse_key_with_one_element(k)
            order_by.append(dict(key=v, op=op))
        elif k == 'limit':
            # TODO check to allow positive integer only if value
            limit = int(v) if v else None
        elif k == 'offset':
            # TODO check to allow positive integer only
            offset = int(v)
        else:
            raise KeyError('Bad querystring : %s=%s' % (k, v))

    return dict(filter_by=filter_by, order_by=order_by, limit=limit,
                offset=offset, tags=tags, context=context)


def base_validator(request, schema, deserializer, only, unknown=INCLUDE):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    logger.debug('base receipt: %r', base)
    if schema is None:
        request.validated = base
        return

    schema = schema(context={'registry': request.anyblok.registry},
                    only=only, unknown=unknown)

    try:
        result = schema.load(base)
        request.validated = result
    except ValidationError as err:
        errors = err.messages
        logger.exception(errors)
        for k, v in errors.items():
            request.errors.add(
                k, 'Validation error for %s' % k,
                v)


def service_collection_get_validator(request, schema=None,
                                     deserializer=None, **kwargs):
    base_validator(request, schema, deserializer, only=['querystring'])


def service_post_validator(request, schema=None, deserializer=None, **kwargs):
    base_validator(request, schema, deserializer, only=['body'])


def service_get_validator(request, schema=None, deserializer=None, **kwargs):
    base_validator(request, schema, deserializer, only=['path'])


def service_put_validator(request, schema=None, deserializer=None, **kwargs):
    base_validator(request, schema, deserializer, only=['path', 'body'])


service_patch_validator = service_put_validator


def service_delete_validator(request, schema=None, deserializer=None, **kwargs):
    base_validator(request, schema, deserializer, only=['path'])


def collection_get_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    deserializer(request)
    # I don't know how validate the query string


def collection_post_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('collection_post', model_name)
    opts = klass.get_deserialize_opts('collection_post')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def collection_patch_validator(request, deserializer=None, klass=None,
                               **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('collection_patch', model_name)
    opts = klass.get_deserialize_opts('collection_patch')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def collection_put_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('collection_put', model_name)
    opts = klass.get_deserialize_opts('collection_put')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def collection_delete_validator(request, deserializer=None, klass=None,
                                **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('collection_delete', model_name)
    opts = klass.get_deserialize_opts('collection_delete')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def get_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the path
    Schema = klass.get_path_schema('get')
    opts = klass.get_path_opts('get')
    klass.apply_validator_schema(request, 'path', Schema, opts, base)


def delete_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the path
    Schema = klass.get_path_schema('get')
    opts = klass.get_path_opts('get')
    klass.apply_validator_schema(request, 'path', Schema, opts, base)


def patch_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the path
    Schema = klass.get_path_schema('patch')
    opts = klass.get_path_opts('patch')
    klass.apply_validator_schema(request, 'path', Schema, opts, base)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('patch', model_name)
    opts = klass.get_deserialize_opts('patch')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def put_validator(request, deserializer=None, klass=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the path
    Schema = klass.get_path_schema('put')
    opts = klass.get_path_opts('put')
    klass.apply_validator_schema(request, 'path', Schema, opts, base)
    # validate the body
    model_name = klass.get_model_name(request, base)
    Schema = klass.get_deserialize_schema('put', model_name)
    opts = klass.get_deserialize_opts('put')
    klass.apply_validator_schema(request, 'body', Schema, opts, base)


def execute_validator(request, deserializer=None, klass=None, schema=None,
                      **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    # validate the path
    Schema = klass.get_path_schema('execute')
    opts = klass.get_path_opts('execute')
    klass.apply_validator_schema(request, 'path', Schema, opts, base)
    # validate the body
    if schema:
        klass.apply_validator_schema(request, 'body', schema, {}, base)
    else:
        request.validated['body'] = base['body']


def collection_execute_validator(request, deserializer=None, klass=None,
                                 schema=None, **kwargs):
    if deserializer is None:
        deserializer = extract_cstruct

    base = deserializer(request)
    if schema:
        klass.apply_validator_schema(request, 'body', schema, {}, base)
    else:
        request.validated['body'] = base['body']



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from uuid import UUID
from datetime import datetime

from pyramid.renderers import JSON

from anyblok.blok import Blok
from anyblok_pyramid.adapter import uuid_adapter, datetime_adapter


class TestBlok6(Blok):

    version = '0.1.0'
    required = ['anyblok-core']

    @classmethod
    def import_declaration_module(cls):
        from . import model # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import model # noqa
        reload(model)

    @classmethod
    def pyramid_load_config(cls, config):
        json_renderer = JSON()
        json_renderer.add_adapter(UUID, uuid_adapter)
        json_renderer.add_adapter(datetime, datetime_adapter)
        config.add_renderer('json', json_renderer)
        config.scan(cls.__module__ + '.views')



# This file is a part of the AnyBlok / Pyramid / REST API project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
# flake8: noqa
from anyblok.config import Configuration
from anyblok import (
        load_init_function_from_entry_points,
        configuration_post_load,
)


load_init_function_from_entry_points(unittest=True)
Configuration.load_config_for_test()


class MockParser:

    def _get_kwargs(self):
        return []

    def _get_args(self):
        return False


Configuration.parse_options(MockParser())
configuration_post_load(unittest=True)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest


class TestCrudServiceBase:
    """Test Service from test_bloks/test_1/views.py:example_service
    This is the basic case, no validators, no schema.
    """

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_1):
        transaction = registry_rest_api_1.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def create_example(self, registry, name="plop"):
        """Create a dummy example record"""
        return registry.Example.insert(name=name)

    def test_example_service_get(self, registry_rest_api_1, webserver):
        """Example GET /anothers/{id}"""
        ex = self.create_example(registry_rest_api_1)
        response = webserver.get('/anothers/%s' % ex.id)
        assert response.status_code == 200
        assert response.json_body.get('name') == "plop"

    def test_example_service_put(self, registry_rest_api_1, webserver):
        """Example PUT /anothers/{id}"""
        ex = self.create_example(registry_rest_api_1)
        response = webserver.put_json(
            '/anothers/%s' % ex.id, {'name': 'plip'})
        assert response.status_code == 200
        assert response.json_body.get('name') == "plip"

    def test_example_service_put_schema_fail_bad_path(
        self, registry_rest_api_1, webserver
    ):
        """Example PUT schema validation fail on invalid path /anothers/{id}
        """
        fail = webserver.put_json(
            '/anothers/x', {'name': 'plip'}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'path'

    def test_example_service_put_schema_fail_bad_value_type(
        self, registry_rest_api_1, webserver
    ):
        """Example PUT schema validation fail on invalid value type in `body`
        /anothers/{id}
        """
        ex = self.create_example(registry_rest_api_1)
        fail = webserver.put_json(
            '/anothers/%s' % ex.id, {'name': 0}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'

    def test_example_service_post(self, registry_rest_api_1, webserver):
        """Example POST /anothers/"""
        response = webserver.post_json('/anothers', {'name': 'plip'})
        assert response.status_code == 200
        assert response.json_body.get('name') == "plip"

    def test_example_service_post_schema_fail_bad_column_name(
        self, registry_rest_api_1, webserver
    ):
        """Example POST schema validation fail on invalid column name /anothers/
        """
        fail = webserver.post_json(
            '/anothers', {'badcolumn': 'plip'}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'

    def test_example_service_post_schema_fail_bad_value_type(
        self, registry_rest_api_1, webserver
    ):
        """Example POST schema validation fail on invalid value type /anothers/
        """
        fail = webserver.post_json(
            '/anothers', {'name': 0}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'


class TestCrudServiceAdvanced:
    """Test Service from test_bloks/test_1/views.py:thing_service
    This is the advanced case, validators and relationnal schema
    """

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_1):
        transaction = registry_rest_api_1.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def create_example(self, registry, name="plop"):
        """Create a dummy example record"""
        return registry.Example.insert(name=name)

    def create_examples(self, registry, examples=None):
        """Create a dummy example collection record"""
        res = []
        if examples is None:
            examples = [dict(name="car example"),
                        dict(name="apple example"),
                        dict(name="human example")]
        for example in examples:
            res.append(self.create_example(registry, **example))
        return res

    def create_thing(self, registry, name="car", secret="aaa", **kwargs):
        """Create a dummy thing record"""
        if 'example_id' not in kwargs.keys():
            example = self.create_example(registry, name="%s example" % name)
            kwargs['example_id'] = example.id
        thing = registry.Thing.insert(name=name, secret=secret, **kwargs)
        return thing

    def create_things(self, registry, things=None):
        """Create a dummy thing collection record"""
        res = []
        if things is None:
            exs = self.create_examples(registry)
            things = [dict(name="car", secret="aaa", example_id=exs[0].id),
                      dict(name="apple", secret="bbb", example_id=exs[1].id),
                      dict(name="human", secret="ccc", example_id=exs[2].id)]
        for thing in things:
            res.append(self.create_thing(registry, **thing))
        return res

    def test_thing_service_get(self, registry_rest_api_1, webserver):
        """Example GET /things/{uuid}"""
        ex = self.create_thing(registry_rest_api_1)
        response = webserver.get('/things/%s' % ex.uuid)
        assert response.status_code == 200
        assert response.json_body.get('name') == "car"

    def test_thing_service_delete(self, registry_rest_api_1, webserver):
        """Example GET /things/{uuid}"""
        ex = self.create_thing(registry_rest_api_1)
        query = registry_rest_api_1.Thing.query().filter_by(uuid=ex.uuid)
        assert query.count() == 1
        response = webserver.delete('/things/%s' % ex.uuid)
        assert response.status_code == 200
        assert query.count() == 0

    def test_thing_service_put(self, registry_rest_api_1, webserver):
        """Example PUT /things/{uuid}"""
        ex = self.create_thing(registry_rest_api_1)
        msg = dict(
            name='plip',
            secret='other',
            example_id=ex.example_id,
        )
        response = webserver.put_json(
            '/things/%s' % ex.uuid, msg)
        assert response.status_code == 200
        assert response.json_body.get('name') == "plip"

    def test_thing_service_put_schema_fail_bad_path(
        self, registry_rest_api_1, webserver
    ):
        """Example PUT schema validation fail on invalid path /things/{uuid}
        """
        fail = webserver.put_json(
            '/things/x', {'name': 'plip'}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') in [
            'path', 'body']

    def test_thing_service_put_schema_fail_bad_value_type(
        self, registry_rest_api_1, webserver
    ):
        """Example PUT schema validation fail on invalid value type in `body`
        /things/{uuid}
        """
        ex = self.create_thing(registry_rest_api_1)
        fail = webserver.put_json(
            '/things/%s' % ex.uuid, {'name': 0}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'

    def test_thing_service_post(self, registry_rest_api_1, webserver):
        """Example POST /things/"""
        ex = self.create_example(registry_rest_api_1)
        response = webserver.post_json(
            '/things',
            {'name': 'book', 'secret': 'xxx', 'example_id': ex.id}
        )
        assert response.status_code == 200
        assert response.json_body.get('name') == "book"

    def test_thing_service_post_schema_fail_bad_column_name(
        self, registry_rest_api_1, webserver
    ):
        """Example POST schema validation fail on invalid column name /things/
        """
        fail = webserver.post_json(
            '/things', {'badcolumn': 'plip'}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'

    def test_thing_service_post_schema_fail_bad_value_type(
        self, registry_rest_api_1, webserver
    ):
        """Example POST schema validation fail on invalid value type /things/
        """
        fail = webserver.post_json(
            '/things', {'name': 0}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'

    def test_thing_collection_get(
        self, registry_rest_api_1, webserver
    ):
        """Thing collection GET /things"""
        self.create_things(registry_rest_api_1)
        response = webserver.get('/things')
        assert response.status_code == 200
        assert len(response.json_body) == 3
        assert response.json_body[0].get('name') == "car"

    def test_thing_collection_get_unknown_querystring(
        self, registry_rest_api_1, webserver
    ):
        """Thing collection GET /things"""
        self.create_things(registry_rest_api_1)
        with pytest.raises(KeyError):
            webserver.get('/things?foo=bar')

    def test_thing_collection_get_with_querystring(
        self, registry_rest_api_1, webserver
    ):
        """Thing collection GET /things?querystring"""
        self.create_things(registry_rest_api_1)
        response = webserver.get('/things?filter[name][like]=car')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 1
        assert int(response.headers.get('X-Count-Records')) == 1
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "car"

    def test_thing_collection_get_with_bad_querystring(
        self, registry_rest_api_1, webserver
    ):
        """Thing collection GET fail on bad /things?badquerystring"""
        self.create_things(registry_rest_api_1)
        fail = webserver.get('/things?filter[name][oops]=car', status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'querystring'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.

"""Setup script for anyblok-pyramid-rest-api"""

from setuptools import setup, find_packages
from os.path import abspath, dirname, join

version = '0.4.0'
here = abspath(dirname(__file__))

with open(join(here, 'README.rst'), 'r',
          encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(join(here, 'CHANGELOG.rst'), 'r',
          encoding='utf-8') as changelog_file:
    changelog = changelog_file.read()

requirements = [
    'anyblok',
    'anyblok_pyramid',
    'cornice',
    'anyblok-marshmallow',
]

test_requirements = [
    # TODO: put package test requirements here
]

anyblok_pyramid_includeme = [
    'pyramid_cornice=anyblok_pyramid_rest_api.pyramid_config:pyramid_cornice',
]
test_bloks = [
    'test_rest_api_1=anyblok_pyramid_rest_api.test_bloks.test_1:TestBlok1',
    'test_rest_api_2=anyblok_pyramid_rest_api.test_bloks.test_2:TestBlok2',
    'test_rest_api_3=anyblok_pyramid_rest_api.test_bloks.test_3:TestBlok3',
    'test_rest_api_4=anyblok_pyramid_rest_api.test_bloks.test_4:TestBlok4',
    'test_rest_api_5=anyblok_pyramid_rest_api.test_bloks.test_5:TestBlok5',
    'test_rest_api_6=anyblok_pyramid_rest_api.test_bloks.test_6:TestBlok6',
    'test_rest_api_7=anyblok_pyramid_rest_api.test_bloks.test_7:TestBlok7',
    'test_rest_api_8=anyblok_pyramid_rest_api.test_bloks.test_8:TestBlok8',
    'test_rest_api_9=anyblok_pyramid_rest_api.test_bloks.test_9:TestBlok9',
    'test_rest_api_10=anyblok_pyramid_rest_api.test_bloks.test_10:TestBlok10',
]

setup(
    name='anyblok_pyramid_rest_api',
    version=version,
    description="Tools to build rest api's",
    long_description=readme + '\n\n' + changelog,
    author="Franck Bret",
    author_email='franckbret@gmail.com',
    url='https://github.com/franckbret/anyblok-pyramid-rest-api',
    packages=find_packages(),
    entry_points={
        'anyblok_pyramid.includeme': anyblok_pyramid_includeme,
        'test_bloks': test_bloks,
    },
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='anyblok-pyramid-rest-api',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: Pyramid',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from datetime import datetime
from uuid import uuid1

from anyblok import Declarations
from anyblok.column import Integer, String, DateTime, Password, UUID
from anyblok.relationship import Many2One


Model = Declarations.Model


@Declarations.register(Model)
class Example():
    """ Example Model for tests purpose
    """
    id = Integer(primary_key=True)
    name = String(label="Name", unique=True, nullable=False)

    def __str__(self):
        return ('{self.name}').format(self=self)

    def __repr__(self):
        msg = ('<Example: {self.name}, {self.id}>')
        return msg.format(self=self)


@Declarations.register(Model)
class Thing():
    uuid = UUID(primary_key=True, default=uuid1, binary=False)
    name = String(label="Name", unique=True, nullable=False)
    create_date = DateTime(default=datetime.now, nullable=False)
    edit_date = DateTime(default=datetime.now, nullable=False)
    secret = Password(
        crypt_context={'schemes': ['pbkdf2_sha512']}, nullable=False)
    example = Many2One(
        label="Example", model=Model.Example, one2many="things",
        nullable=False)

    def __str__(self):
        return ('{self.name}').format(self=self)

    def __repr__(self):
        msg = ('<Thing: {self.name}, {self.uuid}>')
        return msg.format(self=self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_marshmallow import SchemaWrapper
from cornice.resource import view as cornice_view, add_resource, add_view
from cornice import Service
from pyramid.security import Deny, Allow, Everyone, ALL_PERMISSIONS
from pyramid.httpexceptions import HTTPUnauthorized, HTTPNotFound
from anyblok_pyramid_rest_api.querystring import QueryString
from types import MethodType
from .validator import (
    collection_get_validator, collection_post_validator, get_validator,
    delete_validator, put_validator, patch_validator, execute_validator,
    collection_execute_validator, collection_put_validator,
    collection_patch_validator, collection_delete_validator
)
from marshmallow import ValidationError
from contextlib import contextmanager
from logging import getLogger

logger = getLogger(__name__)


@contextmanager
def saved_errors_in_request(request):
    try:
        yield
    except Exception as e:
        logger.exception(e)
        request.errors.add(
            'body', '500 Internal Server Error', str(e))
        request.errors.status = 500
    finally:
        if request.errors:
            logger.debug('Request error found: rollback the registry')
            request.anyblok.registry.rollback()


def get_path(request):
    """Ensure we get a valid path
    """
    return request.validated.get('path', request.matchdict)


def update_from_query_string(request, Model, query, adapter):
    headers = request.response.headers
    if request.params:
        # TODO: Implement schema validation to use request.validated
        querystring = QueryString(request, Model, adapter=adapter)
        total_query = querystring.from_filter_by(query)
        total_query = querystring.from_tags(total_query)
        query = querystring.from_order_by(total_query)
        query = querystring.from_limit(query)
        query = querystring.from_offset(query)
        # TODO: Advanced pagination with Link Header
        # Link: '<https://api.github.com/user/repos?page=3&per_page=100>;
        # rel="next",
        # <https://api.github.com/user/repos?page=50&per_page=100>;
        # rel="last"'
        headers['X-Count-Records'] = str(query.count())
        headers['X-Total-Records'] = str(total_query.count())
        # TODO: Etag / timestamp / 304 if no changes
        # TODO: Cache headers
        return query
    else:
        # no querystring, returns all records (maybe we will want to apply
        # some default filters values
        headers['X-Count-Records'] = str(query.count())
        headers['X-Total-Records'] = str(query.count())
        return query


def post_item(request, Model):
    if not request.errors:
        if isinstance(Model, str):
            Model = request.anyblok.registry.get(Model)

        with saved_errors_in_request(request):
            return Model.insert(**request.validated['body'])


def get_items(request, Model, Adapter=None):
    if not request.errors:
        if isinstance(Model, str):
            Model = request.anyblok.registry.get(Model)

        query = Model.query()
        query = update_from_query_string(request, Model, query, Adapter)
        if not query.count():
            return

        return query.all()


def get_item(request, Model):
    if not request.errors:
        if isinstance(Model, str):
            Model = request.anyblok.registry.get(Model)

        path = get_path(request)
        model_pks = Model.get_primary_keys()
        pks = {x: path[x] for x in model_pks}
        item = Model.from_primary_keys(**pks)
        if item:
            return item
        else:
            path = ', '.join(
                ['%s=%s' % (x, y)
                 for x, y in request.validated.get('path', {}).items()])
            request.errors.add(
                'path', '404 not found',
                'Resource %s with %s does not exist.' % (
                    Model.__registry_name__, path))
            request.errors.status = 404


def put_item(request, Model):
    """
    """
    if not request.errors:
        if isinstance(Model, str):
            Model = request.anyblok.registry.get(Model)

        item = get_item(request, Model)
        if item:
            body = request.validated.get('body', request.validated)
            with saved_errors_in_request(request):
                item.update(**body)

            return item


patch_item = put_item


def delete_item(request, Model):
    """
    """
    if not request.errors:
        if isinstance(Model, str):
            Model = request.anyblok.registry.get(Model)

        item = get_item(request, Model)
        if item:
            with saved_errors_in_request(request):
                item.delete()


def add_execute_on_crud_resource(cls, service_path=None, **kwargs):
    if service_path is None:
        service_path = '%(path)s/execute/%(name)s'

    services = {}
    for attr in dir(cls):
        method = getattr(cls, attr)
        service_kwargs = kwargs.copy()

        # auto-wire klass as its own view factory, unless one
        # is explicitly declared.
        if 'factory' not in kwargs:
            service_kwargs['factory'] = cls

        if getattr(method, 'is_a_crud_resource_execute_on_collection', None):
            service_name = 'collection_' + cls.__name__.lower() + '_execute_'
            service_name += method.crud_resource_execute_name
            service_kwargs['path'] = service_path % {
                'path': service_kwargs.pop('collection_path'),
                'name': method.crud_resource_execute_name,
            }
        elif getattr(method, 'is_a_crud_resource_execute', None):
            service_name = cls.__name__.lower() + '_execute_'
            service_name += method.crud_resource_execute_name
            del service_kwargs['collection_path']
            service_kwargs['path'] = service_path % {
                'path': service_kwargs['path'],
                'name': method.crud_resource_execute_name,
            }
        else:
            continue

        if method.crud_resource_execute_path:
            service_kwargs['path'] = method.crud_resource_execute_path

        service = services[service_name] = Service(
            name=service_name, depth=2, **service_kwargs)
        for view_args in method.__views__:
            verb = view_args.pop('verb', 'POST')
            service.add_view(verb, attr, klass=cls, **view_args)

    cls._services.update(services)
    return cls


def resource(depth=2, **kwargs):

    def wrapper(cls):
        service_path = kwargs.pop('service_path', None)
        klass = add_resource(cls, depth, **kwargs)
        klass = add_execute_on_crud_resource(
            klass, service_path=service_path, **kwargs)
        return klass

    return wrapper


class CrudResource:
    """Main class to define a RESTFUL API on an AnyBlok resource

    ::

        from anyblok_pyramid_rest_api.crud_resource import (
            CrudResource, resource)
        from anyblok_pyramid import current_blok

        @resource(
            collection_path='/myresourcepaths'
            path='/myresourcepath/{my model pk},
            installed_blok=current_blok()
        )
        class MyResource(CrudResource):
            model = 'MyModel'

    This action create view to acces on the AnyBlok Model (MyModel):

    * collection_get: validate the querystring, serialize the output
    * collection_post: deserialize the body input, serialize the output
    * collection_put: validate the querystring, serialize the output
    * get: validate the path, deserialize the body, serialize the output
    * patch: validate the path, deserialize the body, serialize the output
    * put: validate the path, deserialize the body, serialize the output
    * delete: validate the path

    The Serialization and Deserialization is done by the MarshMallow Schema.
    By default the schema is created by AnyBlok MarshMallow, but the schema
    can be overwritten

    * deactivate some access
      - has_collection_get: bool default True
      - has_collection_post: bool default True
      - has_collection_patch: bool default True
      - has_collection_put: bool default True
      - has_collection_delete: bool default True
      - has_get: bool default True
      - has_delete: bool default True
      - has_patch: bool default True
      - has_put: bool default True
    * get schema
      - default_schema: classmethod of AnyBlokMarshmallow schema, by default use
        model to defined it
    * get serialize schema
      - default_serialize_schema: method of AnyBlokMarshmallow schema, by
        default use default_schema
      - serialize_collection_get: method of AnyBlokMarshmallow schema,
                                  by default use default_serialize_schema
      - serialize_collection_post: method of AnyBlokMarshmallow schema
                                   by default use default_serialize_schema
      - serialize_collection_patch: method of AnyBlokMarshmallow schema
                                    by default use default_serialize_schema
      - serialize_collection_put: method of AnyBlokMarshmallow schema
                                  by default use default_serialize_schema
      - serialize_collection_delete: method of AnyBlokMarshmallow schema
                                     by default use default_serialize_schema
      - serialize_get: method of AnyBlokMarshmallow schema, by default use
                       default_serialize_schema
      - serialize_delete: method of AnyBlokMarshmallow schema, by default use
                          default_serialize_schema
      - serialize_patch: method of AnyBlokMarshmallow schema, by default use
                         default_serialize_schema
      - serialize_put: method of AnyBlokMarshmallow schema, by default use
                       default_serialize_schema
    * get serialize opts
      - get_serialize_opts: method return dict of option to use
    * get deserialize schema
      - default_deserialize_schema: method of AnyBlokMarshmallow schema, by
                                    default use default_schema
      - deserialize_collection_post: method of AnyBlokMarshmallow schema,
                                     by default use default_deserialize_schema
      - deserialize_collection_patch: method of AnyBlokMarshmallow schema,
                                      by default use default_deserialize_schema
      - deserialize_collection_put: method of AnyBlokMarshmallow schema,
                                    by default use default_deserialize_schema
      - deserialize_collection_delete: method of AnyBlokMarshmallow schema,
                                       by default use default_deserialize_schema
      - deserialize_patch: method of AnyBlokMarshmallow schema,
                           by default use default_deserialize_schema
      - deserialize_put: method of AnyBlokMarshmallow schema,
                         by default use default_deserialize_schema
    * get deserialize opts
      - get_deserialize_opts: method return dict of option to use
    * get path schema
      - default_path_schema: method of AnyBlokMarshmallow schema, by default use
                             default_schema
      - path_get: method of AnyBlokMarshmallow schema, by default use
                  default_serialize_schema
      - path_delete: method of AnyBlokMarshmallow schema, by default use
                     default_serialize_schema
      - path_patch: method of AnyBlokMarshmallow schema, by default use
                    default_serialize_schema
      - path_put: method of AnyBlokMarshmallow schema, by default use
                  default_serialize_schema
      - path_execute: method of AnyBlokMarshmallow schema, by default use
                      default_serialize_schema
    * get path opts
      - get_path_opts: method return dict of option to use
    * update_collection_get_filter: method to improve query to filter
    * create
    * update
    * collection_update
    * delete_entry
    * delete_entries
    * get_model_name:

    You can update the querystring to add an adapter in the resource::

        @resource(...)
        class MyResource(CrudResource):
            ...

            class adapter_cls(Adapter):
                # See the adapter definition

    Some http post action can be added with the decorator execute::

        @resource(...)
        class MyResource(CrudResource):
            ...

            @CrudResource.execute('action_name')  # schema is optional
            def post_foo_bar(self):
                # path/execute/action_name
                querystring = self.get_querystring('action_name')
                body = self.body
                ...

            @CrudResource.execute('other_action_name',
                                  collection=True)  # schema is optional
            def post_foo_bar(self):
                # collection_path/execute/other_action_name
                body = self.body
                Model = self.get_model('other_action_name')
                item = get_item(self.request, Model)
                ...
    """

    model = None
    adapter_cls = None
    cache_default_schema = True  # TODO
    has_collection_get = True
    has_collection_post = True
    has_collection_patch = True
    has_collection_put = True
    has_collection_delete = True
    has_get = True
    has_delete = True
    has_patch = True
    has_put = True

    def __init__(self, request, **kwargs):
        self.request = request
        self.registry = self.request.anyblok.registry
        self.adapter = None
        if self.adapter_cls:
            self.adapter = self.adapter_cls(
                self.registry,
                Model=self.get_model('collection_get'))
            self.adapter.load_decorators()

    @classmethod
    def get_model_name(cls, request, rest_action=None, base=None):
        return cls.model

    def view_is_activated(self, condition):
        if not condition:
            raise HTTPNotFound()

    def model_name(self, rest_action=None):
        return self.get_model_name(
            self.request,
            rest_action=rest_action,
            base=self.request.validated
        )

    def __acl__(self):
        if not hasattr(self, 'registry'):
            raise HTTPUnauthorized("ACL have not get AnyBlok registry")

        Blok = self.registry.System.Blok
        if not Blok.is_installed('auth'):
            return [(Allow, Everyone, ALL_PERMISSIONS)]

        userid = self.request.authenticated_userid
        if userid:
            return self.registry.User.get_acl(
                userid, self.model_name(), params=dict(self.request.matchdict)
            )

        return [(Deny, Everyone, ALL_PERMISSIONS)]

    @classmethod
    def apply_validator_schema(cls, request, part, Schema, opts, base):
        if 'context' not in opts:
            opts['context'] = {}

        opts['context']['registry'] = request.anyblok.registry
        try:
            logger.debug('Validate %r with schema %r and option %r',
                         base[part], Schema, opts)
            schema = Schema(**opts)
            result = schema.load(base[part])
            request.validated[part] = result
        except ValidationError as err:
            request.anyblok.registry.rollback()
            logger.exception(err)
            errors = err.messages
            for k, v in errors.items():
                request.errors.add(
                    part, 'Validation error for %s' % part,
                    {k: v})

    def get_model(self, rest_action):
        return self.registry.get(self.model_name(rest_action=rest_action))

    def update_collection_get_filter(self, query):
        return query

    @classmethod
    def get_value_or_call_method(cls, attribute, *args, **kwargs):
        value = None
        if hasattr(cls, attribute):
            value = getattr(cls, attribute)
            if isinstance(value, MethodType):
                value = value(*args, **kwargs)

        return value

    @classmethod
    def default_schema(cls, name, model_name):
        name = '%s.%s.%s' % (cls.__name__, name, model_name)
        return type(name, (SchemaWrapper,), {'model': model_name})

    @classmethod
    def default_serialize_schema(cls, model_name):
        return cls.get_value_or_call_method(
            'default_schema', 'Serialize', model_name)

    @classmethod
    def get_serialize_schema(cls, rest_action, model_name):
        Schema = cls.get_value_or_call_method(
            'serialize_' + rest_action, model_name)
        if Schema is None:
            Schema = cls.get_value_or_call_method(
                'default_serialize_schema', model_name)

        return Schema

    @classmethod
    def get_serialize_opts(cls, rest_action):
        opts = {}
        if rest_action in ('collection_get',
                           'collection_post',
                           'collection_put',
                           'collection_patch'):
            opts['many'] = True

        return opts

    @classmethod
    def default_deserialize_schema(cls, model_name):
        return cls.get_value_or_call_method(
            'default_schema', 'Deserialize', model_name)

    @classmethod
    def get_deserialize_schema(cls, rest_action, model_name):
        Schema = cls.get_value_or_call_method(
            'deserialize_' + rest_action, model_name)
        if Schema is None:
            Schema = cls.get_value_or_call_method(
                'default_deserialize_schema', model_name)

        return Schema

    @classmethod
    def get_deserialize_opts(cls, rest_action):
        opts = {}
        if rest_action == 'collection_patch':
            opts['partial'] = True
            opts['many'] = True
        elif rest_action in ('collection_put', 'collection_post'):
            opts['many'] = True
        elif rest_action == 'collection_delete':
            opts['many'] = True
            opts['context'] = {'only_primary_key': True}
        elif rest_action == 'patch':
            opts['partial'] = True

        return opts

    @classmethod
    def default_path_schema(cls):
        return cls.get_value_or_call_method('default_schema', 'Path', cls.model)

    @classmethod
    def get_path_schema(cls, rest_action):
        Schema = cls.get_value_or_call_method('path_' + rest_action)
        if Schema is None:
            Schema = cls.get_value_or_call_method('default_path_schema')

        return Schema

    @classmethod
    def get_path_opts(cls, rest_action):
        opts = {'context': {'only_primary_key': True}}
        return opts

    def serialize(self, rest_action, entry):
        Schema = self.get_serialize_schema(
            rest_action, self.model_name(rest_action))
        opts = self.get_serialize_opts(rest_action)
        opts['context'] = {'registry': self.registry}
        schema = Schema(**opts)
        return schema.dump(entry)

    @property
    def body(self):
        return self.request.validated.get('body', self.request.validated)

    def get_querystring(self, rest_action):
        Model = self.get_model(rest_action)
        query = self.update_collection_get_filter(Model.query())
        query = update_from_query_string(
            self.request, Model, query, self.adapter)
        return query

    @cornice_view(validators=(collection_get_validator,), permission="read")
    def collection_get(self):
        self.view_is_activated(self.has_collection_get)
        if not self.request.errors:
            query = self.get_querystring('collection_get')
            if not query.count():
                return []

            return self.serialize('collection_get', query.all())

    def create(self, Model, params):
        return Model.insert(**params)

    @cornice_view(validators=(collection_post_validator,), permission="create")
    def collection_post(self):
        self.view_is_activated(self.has_collection_post)
        if not self.request.errors:
            items = []
            Model = self.get_model('collection_post')
            for params in self.body:
                with saved_errors_in_request(self.request):
                    items.append(self.create(Model, params=params))

            if items and not self.request.errors:
                return self.serialize('collection_post', items)

    def collection_update(self, Model, body):
        items = []
        for params in body:
            try:
                pks = {x: params[x] for x in Model.get_primary_keys()}
            except KeyError as e:
                self.request.errors.add(
                    'body', 'Validation Error',
                    'No primary key found %r to get the item on %s' % (
                        e.args, Model))
                self.request.errors.status = 400
            else:
                item = Model.from_primary_keys(**pks)
                if item:
                    self.update(item, params=params)
                    items.append(item)
                else:
                    self.request.errors.add(
                        'body', 'Validation Error',
                        'The primary key found %r does not exist on %s' % (
                            pks, Model))
                    self.request.errors.status = 400

        return items

    @cornice_view(validators=(collection_patch_validator,), permission="update")
    def collection_patch(self):
        self.view_is_activated(self.has_collection_patch)
        if not self.request.errors:
            items = []
            Model = self.get_model('collection_patch')
            with saved_errors_in_request(self.request):
                items = self.collection_update(Model, self.body)

            return self.serialize('collection_patch', items)

    @cornice_view(validators=(collection_put_validator,), permission="update")
    def collection_put(self):
        self.view_is_activated(self.has_collection_put)
        if not self.request.errors:
            items = []
            Model = self.get_model('collection_put')
            with saved_errors_in_request(self.request):
                items = self.collection_update(Model, self.body)

            return self.serialize('collection_put', items)

    def delete_entries(self, Model, body):
        for params in body:
            try:
                pks = {x: params[x] for x in Model.get_primary_keys()}
            except KeyError as e:
                self.request.errors.add(
                    'body', 'Validation Error',
                    'No primary key found %r to get the item on %s' % (
                        e.args, Model))
                self.request.errors.status = 400
                return 0
            else:
                item = Model.from_primary_keys(**pks)
                if item:
                    self.delete_entry(item)
                else:
                    self.request.errors.add(
                        'body', 'Validation Error',
                        'The primary key found %r does not exist on %s' % (
                            pks, Model))
                    self.request.errors.status = 400
                    return 0

        return len(body)

    @cornice_view(validators=(collection_delete_validator,),
                  permission="delete")
    def collection_delete(self):
        self.view_is_activated(self.has_collection_delete)
        count = 0
        if not self.request.errors:
            Model = self.get_model('collection_delete')
            with saved_errors_in_request(self.request):
                count = self.delete_entries(Model, self.body)

        return count

    @cornice_view(validators=(get_validator,), permission="read")
    def get(self):
        self.view_is_activated(self.has_get)
        if not self.request.errors:
            Model = self.get_model('get')
            item = get_item(self.request, Model)
            if item:
                return self.serialize('get', item)

    def delete_entry(self, item):
        item.delete()

    @cornice_view(validators=(delete_validator,), permission="delete")
    def delete(self):
        """
        """
        self.view_is_activated(self.has_delete)
        if not self.request.errors:
            Model = self.get_model('delete')
            item = get_item(self.request, Model)
            if item:
                with saved_errors_in_request(self.request):
                    self.delete_entry(item)

            return {}

    def update(self, item, params=None):
        if params:
            item.update(**params)

    @cornice_view(validators=(patch_validator,), permission="update")
    def patch(self):
        """
        """
        self.view_is_activated(self.has_patch)
        if not self.request.errors:
            Model = self.get_model('patch')
            item = get_item(self.request, Model)
            if item:
                with saved_errors_in_request(self.request):
                    self.update(item, params=self.body)

                return self.serialize('patch', item)

    @cornice_view(validators=(put_validator,), permission="update")
    def put(self):
        """
        """
        self.view_is_activated(self.has_put)
        if not self.request.errors:
            Model = self.get_model('put')
            item = get_item(self.request, Model)
            if item:
                with saved_errors_in_request(self.request):
                    self.update(item, params=self.body)

                return self.serialize('put', item)

    @classmethod
    def service(cls, name, permission=None, collection=False, path=None,
                **kwargs):
        if permission is None:
            permission = name

        def wrapper(method):
            method.crud_resource_execute_name = name
            method.crud_resource_execute_path = path

            if collection:
                method.is_a_crud_resource_execute_on_collection = True
                validators = (collection_execute_validator,)
            else:
                method.is_a_crud_resource_execute = True
                validators = (execute_validator,)

            add_view(method, validators=validators, permission=permission,
                     **kwargs)
            return method

        return wrapper



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.tests.testcase import LogCapture


class TestCrudResourceBase:
    """Test CrudResource class from test_bloks/test_1/views.py:ExampleResource
    This is the basic case, no validators(except the default one,
    base_validator), no schema.
    """

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_1, webserver):
        transaction = registry_rest_api_1.begin_nested()
        self.registry = registry_rest_api_1

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def test_example_get(self):
        """Example GET /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.get('/examples/%s' % ex.id)
        assert response.status_code == 200
        assert response.json_body.get('name') == "plop"

    def test_example_get_bad_value_in_path(self):
        """Example FAILED GET /examples/{id}"""
        self.create_example()
        fail = self.webserver.get('/examples/0', status=404)
        assert fail.status_code == 404

    def test_example_collection_post(self):
        """Example POST /examples/"""
        response = self.webserver.post_json('/examples', [{'name': 'plip'}])
        assert response.status_code == 200
        assert response.json_body[0]['name'] == "plip"

    def test_example_collection_post_with_errors(self):
        """Example POST /examples/with/errors"""
        with LogCapture() as logs:
            fail = self.webserver.post_json(
                '/examples/with/errors', [{'name': 'plip'}], status=500)

        assert fail.status_code == 500
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'
        assert fail.json_body.get('errors')[0].get('description') == 'test'

        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_put(self):
        """Example PUT /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.put_json(
            '/examples/%s' % ex.id, {'name': 'plip'})
        assert response.status_code == 200
        assert response.json_body.get('name') == "plip"

    def test_example_put_with_errors(self):
        """Example PUT /examples/with/error/{id}"""
        ex = self.create_example()
        with LogCapture() as logs:
            self.webserver.put_json(
                '/examples/with/error/%s' % ex.id,
                {'name': 'plip'}, status=500)

        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_patch_with_errors(self):
        """Example PATCH /examples/with/error/{id}"""
        ex = self.create_example()
        with LogCapture() as logs:
            self.webserver.patch_json(
                '/examples/with/error/%s' % ex.id, {'name': 'plip'},
                status=500)

        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_collection_put_with_errors(self):
        """Example PUT /examples/with/errors"""
        ex = self.create_example()
        with LogCapture() as logs:
            response = self.webserver.put_json(
                '/examples/with/errors',
                [{'id': ex.id, 'name': 'plip'}], status=500)

        assert response.status_code == 500
        assert ex.name != "plip"
        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_collection_patch_with_errors(self):
        """Example PATCH /examples/with/errors"""
        ex = self.create_example()
        with LogCapture() as logs:
            response = self.webserver.patch_json(
                '/examples/with/errors',
                [{'id': ex.id, 'name': 'plip'}], status=500)

        assert response.status_code == 500
        assert ex.name != "plip"
        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_put_bad_value_in_path(self):
        """Example FAILED PUT /examples/{id}"""
        self.create_example()
        fail = self.webserver.put_json(
            '/examples/0', {'name': 'plip'}, status=404)
        assert fail.status_code == 404

    def test_example_delete(self):
        """Example DELETE /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.delete('/examples/%s' % ex.id)
        assert response.status_code == 200
        response = self.webserver.get('/examples')
        assert response.status_code == 200
        assert response.json_body == []

    def test_example_delete_with_errors(self):
        """Example DELETE /examples/with/error/{id}"""
        ex = self.create_example()
        with LogCapture() as logs:
            response = self.webserver.delete(
                '/examples/with/error/%s' % ex.id, status=500)

        assert response.status_code == 500
        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_collection_delete_with_errors(self):
        """Example DELETE /examples/with/errors"""
        ex = self.create_example()
        with LogCapture() as logs:
            response = self.webserver.delete_json(
                '/examples/with/errors',
                [{'id': ex.id}], status=500)

        assert response.status_code == 500
        assert ('Request error found: rollback the registry'
                in logs.get_debug_messages())

    def test_example_delete_bad_value_in_path(self):
        """Example FAILED DELETE /examples/{id}"""
        self.create_example()
        fail = self.webserver.delete('/examples/0', status=404)
        assert fail.status_code == 404

    def test_example_collection_get(self):
        """Example collection GET /examples"""
        self.create_example()
        response = self.webserver.get('/examples')
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "plop"

    def test_example_collection_get_order_by_asc(self):
        """Example collection GET /examples?order_by[asc]=field"""
        names = ['air', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?order_by[asc]=name')
        assert response.status_code == 200
        assert len(response.json_body) == 2
        assert response.json_body[0].get('name') == "air"

    def test_example_collection_get_order_by_desc(self):
        """Example collection GET /examples?order_by[desc]=field"""
        names = ['air', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?order_by[desc]=name')
        assert response.status_code == 200
        assert len(response.json_body) == 2
        assert response.json_body[0].get('name') == "zen"

    def test_example_collection_get_filter_eq(self):
        """Example collection GET /examples?filter[name][eq]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][eq]=air')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 1
        assert int(response.headers.get('X-Count-Records')) == 1
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "air"

    def test_example_collection_get_not_filter_eq(self):
        """Example collection GET /examples?~filter[name][eq]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?~filter[name][eq]=air')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 4
        assert int(response.headers.get('X-Count-Records')) == 4
        assert len(response.json_body) == 4

    def test_example_collection_get_filter_like(self):
        """Example collection GET /examples?filter[name][like]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][like]=a')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 3
        assert int(response.headers.get('X-Count-Records')) == 3
        assert len(response.json_body) == 3
        assert response.json_body[0].get('name') == "air"

    def test_example_collection_get_filter_ilike(self):
        """Example collection GET /examples?filter[name][ilike]=TeRm"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][ilike]=A')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 3
        assert int(response.headers.get('X-Count-Records')) == 3
        assert len(response.json_body) == 3
        assert response.json_body[0].get('name') == "air"

    def test_example_collection_get_limit(self):
        """Example collection GET /examples?limit=limit"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?limit=4')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 5
        assert len(response.json_body) == 4
        assert response.json_body[0].get('name') == "air"

    def test_example_collection_get_offset(self):
        """Example collection GET /examples?offset=offset"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?offset=3')
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 5
        assert len(response.json_body) == 2
        assert response.json_body[0].get('name') == "dot"


class TestCrudResourceWithDefaultSchema:
    """Test CrudResource class from
    test_bloks/test_1/views.py:ExampleResourceBaseValidator.

    This is the basic case, no validators(except the default one,
    base_validator), no schema.
    """

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_1, webserver):
        transaction = registry_rest_api_1.begin_nested()
        self.registry = registry_rest_api_1

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def test_example_get(self):
        """Example GET /with/default/schema/examples/{id}"""
        ex = self.create_example()
        response = self.webserver.get(
            '/with/default/schema/examples/%s' % ex.id)
        assert response.status_code == 200
        assert response.json_body.get('name') == "plop"

    def test_example_get_bad_value_in_path(self):
        """Example FAILED GET /with/default/schema/examples/{id}"""
        self.create_example()
        fail = self.webserver.get('/with/default/schema/examples/0', status=404)
        assert fail.status_code == 404


class CrudResourceSchema:

    def create_customer(self, name="bob", tag_name="green", zipcode="000"):
        """Create a dummy customer record"""
        tag = self.registry.Tag.insert(name=tag_name)
        customer = self.registry.Customer.insert(name=name)
        customer.tags.append(tag)
        city = self.registry.City.insert(name="nowhere", zipcode=zipcode)
        self.registry.Address.insert(
            customer=customer, city=city, street="Dead end street")
        return customer

    def test_example_collection_get(self):
        """Example collection GET /examples"""
        self.create_customer()
        response = self.webserver.get(self.collection_path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "bob"
        assert len(response.json_body[0].get('addresses')) == 1
        assert response.json_body[0].get('addresses')[0].get('city').get(
            'zipcode') == "000"

    def test_customer_get(self):
        """Customer GET /customers/{id}"""
        cu = self.create_customer()
        response = self.webserver.get(self.path % cu.id)
        assert response.status_code == 200
        assert response.json_body.get('name') == "bob"
        assert len(response.json_body.get('addresses')) == 1
        assert response.json_body.get('addresses')[0].get('city').get(
            'zipcode') == "000"

    def test_customer_get_bad_value_in_path(self):
        """Customer FAILED GET /customers/{id}"""
        fail = self.webserver.get(self.path % '0', status=404)
        assert fail.status_code == 404

    def test_customer_collection_post(self):
        """Customer POST /customers"""
        response = self.webserver.post_json(self.collection_path,
                                            [{'name': 'plip'}])
        assert response.status_code == 200
        assert response.json_body[0]['name'] == "plip"

    def test_customer_collection_post_partial(self):
        """Customer POST partial /customers"""
        response = self.webserver.post_json(self.collection_path,
                                            [{'name': 'plip'}])
        assert response.status_code == 200
        assert response.json_body[0]['name'] == "plip"

    def test_customer_collection_post_empty_body(self):
        """Customer POST empty body /customers"""
        fail = self.webserver.post_json(
            self.collection_path, [{}], status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'
        assert fail.json_body.get('errors')[0].get('description') == {
            '0': {'name': ['Missing data for required field.']}}

    def test_customer_collection_post_bad_key_in_body(self):
        """Customer POST bad key in body /customers"""
        fail = self.webserver.post_json(
            self.collection_path, [{'unexistingkey': 'plip'}], status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('status') == 'error'
        assert fail.json_body.get('errors')[0].get('location') == 'body'
        assert fail.json_body.get('errors')[0].get(
            'name') == 'Validation error for body'
        assert fail.json_body['errors'][0]['description'] == {
            '0': {
                'name': ['Missing data for required field.'],
                'unexistingkey': [
                    'Unknown field.',
                    "Unknown fields {'unexistingkey'} on Model "
                    "Model.Customer"
                ]
            }
        }

    def test_customer_put(self):
        """Customer PUT /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.head(self.path % ex.id)  # fix headers
        response = self.webserver.put_json(
            self.path % ex.id, {'name': 'bobby'})
        assert response.status_code == 200
        assert response.json_body.get('name') == "bobby"

    def test_customer_patch(self):
        """Customer PATCH /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.head(self.path % ex.id)  # fix headers
        response = self.webserver.patch_json(
            self.path % ex.id, {'name': 'bobby'})
        assert response.status_code == 200
        assert response.json_body.get('name') == "bobby"

    def test_customer_put_bad_value_in_path(self):
        """Customer FAILED PUT /customers/{id}"""
        self.create_customer()
        fail = self.webserver.put_json(
            self.path % '0', {'name': 'plip'}, status=404)
        assert fail.status_code == 404

    def test_customer_delete(self):
        """Customer DELETE /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.delete(self.path % ex.id)
        assert response.status_code == 200
        response = self.webserver.get(self.collection_path)
        assert response.status_code == 200
        assert response.json_body == []

    def test_customer_delete_bad_value_in_path(self):
        """Customer FAILED DELETE /customers/{id}"""
        fail = self.webserver.get(self.path % '0', status=404)
        assert fail.status_code == 404

    def test_customer_collection_put(self):
        """Customer PUT /customers"""
        ex = self.create_customer()
        response = self.webserver.head(self.collection_path)  # fix headers
        response = self.webserver.put_json(
            self.collection_path, [{'id': ex.id, 'name': 'bobby'}])
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "bobby"

    def test_customer_collection_patch(self):
        """Customer PATCH /customers"""
        ex = self.create_customer()
        response = self.webserver.head(self.collection_path)  # fix headers
        response = self.webserver.patch_json(
            self.collection_path, [{'id': ex.id, 'name': 'bobby'}])
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "bobby"

    def test_customer_collection_put_no_entry_found(self):
        """Customer FAILED PUT /customers (no id)"""
        response = self.webserver.put_json(
            self.collection_path, [{'name': 'plip'}], status=400)
        assert response.status_code == 400

    def test_customer_collection_put_no_entry_found2(self):
        """Customer FAILED PUT /customers (id=0)"""
        response = self.webserver.put_json(
            self.collection_path, [{'id': 0, 'name': 'plip'}], status=400)
        assert response.status_code == 400

    def test_customer_collection_delete(self):
        """Customer DELETE /customers"""
        ex = self.create_customer()
        path = self.collection_path + '?filter[id][eq]=%d' % ex.id

        response = self.webserver.get(path)
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 1
        assert int(response.headers.get('X-Count-Records')) == 1

        response = self.webserver.delete_json(
            self.collection_path, [{'id': ex.id}])
        assert response.status_code == 200
        assert response.json_body == 1

        response = self.webserver.get(path)
        assert response.status_code == 200
        assert int(response.headers.get('X-Total-Records')) == 0
        assert int(response.headers.get('X-Count-Records')) == 0

    def test_customer_collection_delete_no_entry_found(self):
        """Customer FAILED DELETE /customers (no id)"""
        response = self.webserver.delete_json(
            self.collection_path, [{}], status=400)
        assert response.status_code == 400

    def test_customer_collection_delete_no_entry_found2(self):
        """Customer FAILED DELETE /customers (id=0)"""
        response = self.webserver.delete_json(
            self.collection_path, [{'id': 0}], status=400)
        assert response.status_code == 400


class TestCrudResourceModelSchema(CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_3/views.py
    """
    collection_path = '/customers/v3'
    path = '/customers/v3/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_3, webserver):
        transaction = registry_rest_api_3.begin_nested()
        self.registry = registry_rest_api_3

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class CrudResourceAdapter:

    def create_adapter_customers(self):
        self.create_customer()
        self.create_customer(name="robert", tag_name="orange", zipcode="001")

    def test_adapter_get_all_without_tag_or_custom_filter(self):
        self.create_adapter_customers()
        response = self.webserver.get(self.collection_path)
        assert response.status_code == 200
        assert len(response.json_body) == 2

    def test_adapter_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=green"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "bob"

    def test_adapter_unexisting_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=unexisting"
        response = self.webserver.get(path, status=400)
        assert response.status_code == 400

    def test_adapter_wrong_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=wrong"
        response = self.webserver.get(path, status=400)
        assert response.status_code == 400
        assert len(response.json_body) == 2

    def test_adapter_tags_1(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=green"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "bob"

    def test_adapter_tags_2(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=green,orange"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 0

    def test_adapter_tags_3(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=orange"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "robert"

    def test_adapter_tags_and_context_1(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=colors&context[colors]=orange"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "robert"

    def test_adapter_tags_and_context_2(self):
        self.create_adapter_customers()
        path = self.collection_path
        path += "?tag=colors&context[colors]=green,orange"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 2

    def test_adapter_customer_filter(self):
        self.create_adapter_customers()
        path = self.collection_path + "?filter[addresses.city][ilike]=001"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert len(response.json_body) == 1
        assert response.json_body[0].get('name') == "robert"

    def test_adapter_wrong_filter(self):
        self.create_adapter_customers()
        path = self.collection_path + "?filter[addresses.other][ilike]=001"
        response = self.webserver.get(path, status=400)
        assert response.status_code == 400
        assert len(response.json_body) == 2

    def test_adapter_wrong_order_by(self):
        self.create_adapter_customers()
        path = self.collection_path + "?order_by[asc]=wrong"
        response = self.webserver.get(path, status=400)
        assert response.status_code == 400
        assert len(response.json_body) == 2

    def test_adapter_order_by(self):
        self.create_adapter_customers()
        path = self.collection_path + "?order_by[asc]=address"
        response = self.webserver.get(path)
        assert response.status_code == 200
        assert response.json_body[0].get('name') == "bob"


class TestCrudResourceWithAdapter(CrudResourceAdapter, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_4/views.py
    """
    collection_path = '/customers/v4'
    path = '/customers/v4/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_4, webserver):
        transaction = registry_rest_api_4.begin_nested()
        self.registry = registry_rest_api_4

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceWithAdapter2(CrudResourceAdapter, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_6/views.py
    """
    collection_path = '/customers/v6'
    path = '/customers/v6/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_6, webserver):
        transaction = registry_rest_api_6.begin_nested()
        self.registry = registry_rest_api_6

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceModelSchemaValidator(CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """
    collection_path = '/customers/v5'
    path = '/customers/v5/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_5, webserver):
        transaction = registry_rest_api_5.begin_nested()
        self.registry = registry_rest_api_5

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class CrudResourceAction:

    def create_customer(self, name="bob", tag_name="green", zipcode="000"):
        """Create a dummy customer record"""
        tag = self.registry.Tag.insert(name=tag_name)
        customer = self.registry.Customer.insert(name=name)
        customer.tags.append(tag)
        city = self.registry.City.insert(name="nowhere", zipcode=zipcode)
        self.registry.Address.insert(
            customer=customer, city=city, street="Dead end street")
        return customer

    def test_collection_action(self):
        """Example collection GET /customers/v*/execute/action1"""
        response = self.webserver.get(self.collection_service_path)
        assert response.status_code == 200
        assert response.json_body == 'test'

    def test_collection_action_post_on_path(self):
        """Example collection POST /customers/v*/{id}/execute/action1"""
        cu = self.create_customer()
        service_path = self.path + '/execute/action1'
        fail = self.webserver.get(service_path % cu.id, status=404)
        assert fail.status_code == 404

    def test_collection_action_post(self):
        """Example collection GET /customers/v*/execute/action1"""
        fail = self.webserver.post(self.collection_service_path, status=405)
        assert fail.status_code == 405

    def test_collection_action_put(self):
        """Example collection PUT /customers/v*/execute/action1"""
        fail = self.webserver.put(self.collection_service_path, status=405)
        assert fail.status_code == 405

    def test_collection_action_patch(self):
        """Example collection PATCH /customers/v*/execute/action1"""
        fail = self.webserver.patch(self.collection_service_path, status=405)
        assert fail.status_code == 405

    def test_collection_action_delete(self):
        """Example collection DELETE /customers/v*/execute/action1"""
        fail = self.webserver.delete(self.collection_service_path, status=405)
        assert fail.status_code == 405

    def test_action(self):
        """Example POST /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        response = self.webserver.post_json(
            self.service_path % cu.id, {'name': 'test'})
        assert response.status_code == 200
        assert response.json_body == 'test'

    def test_action_missing_name(self):
        """Example FAIL POST /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        fail = self.webserver.post_json(
            self.service_path % cu.id, {}, status=400)
        assert fail.status_code == 400
        assert fail.json_body.get('errors')[0].get('location') == 'body'
        assert fail.json_body.get('errors')[0].get(
            'name') == 'Validation error for body'

    def test_action_post_on_collection_path(self):
        """Example collection POST /customers/v*/execute/action2"""
        path = self.collection_path + '/execute/action2'
        fail = self.webserver.get(path, status=404)
        assert fail.status_code == 404

    def test_action_get(self):
        """Example GET /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        fail = self.webserver.get(self.service_path % cu.id, status=405)
        assert fail.status_code == 405

    def test_action_put(self):
        """Example PUT /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        fail = self.webserver.put(self.service_path % cu.id, status=405)
        assert fail.status_code == 405

    def test_action_patch(self):
        """Example PATCH /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        fail = self.webserver.patch(self.service_path % cu.id, status=405)
        assert fail.status_code == 405

    def test_action_delete(self):
        """Example DELETE /customers/v*/{id}/execute/action2"""
        cu = self.create_customer()
        fail = self.webserver.delete(self.service_path % cu.id, status=405)
        assert fail.status_code == 405


class TestCrudResourceExecute1(CrudResourceAction, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_7/views.py
    """
    collection_path = '/customers/v7'
    path = '/customers/v7/%s'
    collection_service_path = '/customers/v7/execute/action1'
    service_path = '/customers/v7/%s/execute/action2'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_7, webserver):
        transaction = registry_rest_api_7.begin_nested()
        self.registry = registry_rest_api_7

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceExecute2(CrudResourceAction, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_8/views.py
    """
    collection_path = '/customers/v8'
    path = '/customers/v8/%s'
    collection_service_path = '/customers/v8/execute/action1'
    service_path = '/customers/v8/%s/execute/action2'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_8, webserver):
        transaction = registry_rest_api_8.begin_nested()
        self.registry = registry_rest_api_8

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceExecute3(CrudResourceAction, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_9/views.py
    """
    collection_path = '/customers/v9'
    path = '/customers/v9/%s'
    collection_service_path = '/customers/v9/other/action1'
    service_path = '/customers/v9/%s/other/action2'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_9, webserver):
        transaction = registry_rest_api_9.begin_nested()
        self.registry = registry_rest_api_9

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceExecute4(CrudResourceAction, CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_10/views.py
    """
    collection_path = '/customers/v10'
    path = '/customers/v10/%s'
    collection_service_path = '/foo/bar'
    service_path = '/customers/v10/%s/execute/action2'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_10, webserver):
        transaction = registry_rest_api_10.begin_nested()
        self.registry = registry_rest_api_10

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        return


class TestCrudResourceModelSchemaValidatorAndAuth(CrudResourceSchema):
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """
    collection_path = '/customers/v5'
    path = '/customers/v5/%s'

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_rest_api_5_logged, webserver):
        transaction = registry_rest_api_5_logged.begin_nested()
        self.registry = registry_rest_api_5_logged

        def rollback():
            try:
                transaction.rollback()
            except Exception:
                pass

        request.addfinalizer(rollback)
        self.webserver = webserver
        webserver.post_json(
            '/login', {'login': 'jssuzanne', 'password': 'mypassword'},
            status=302)
        return

    def test_customer_get_logout(self):
        """Customer GET /customers/{id} [UNLOGGED]"""
        cu = self.create_customer()
        self.webserver.post_json('/logout', {}, status=302)
        self.webserver.get(self.path % cu.id, status=403)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from cornice import Service
from cornice.resource import resource
from anyblok_pyramid import current_blok
from anyblok_pyramid_rest_api.validator import (
    service_collection_get_validator,
    service_post_validator,
    service_get_validator,
    service_put_validator,
    service_delete_validator,
)
from anyblok_pyramid_rest_api.crud_resource import (
    CrudResource,
    get_items,
    post_item,
    get_item,
    put_item,
    delete_item,
)
from .schema import (ExampleSchema, ExamplePathSchema, ThingSchema,
                     ThingRequestSchema, AnotherSchema)


@resource(collection_path='/examples', path='/examples/{id}',
          installed_blok=current_blok())
class ExampleResource(CrudResource):
    model = 'Model.Example'


@resource(collection_path='/with/default/schema/examples',
          path='/with/default/schema/examples/{id}',
          installed_blok=current_blok())
class ExampleResourceWithDefaultSchema(CrudResource):
    model = 'Model.Example'
    default_schema = ExampleSchema
    default_path_schema = ExamplePathSchema


# another endpoint through a service with the same model
another_service = Service(name='another_service', path='/anothers/{id}')


# another collection endpoint through a service with the same model
another_collection_service = Service(
    name='another_collection_service',
    path='/anothers')


@another_service.get()
def another_service_get(request):
    """ No validator, no schema
    """
    registry = request.anyblok.registry
    model = registry.get('Model.Example')
    item = model.query().get(request.matchdict['id'])
    return item.to_dict()


@another_service.put(validators=(service_put_validator,), schema=AnotherSchema)
def another_service_put(request):
    """ full_validator + AnotherSchema
    """
    registry = request.anyblok.registry
    Model = registry.get('Model.Example')
    item = Model.query().get(request.validated.get('path').get('id'))
    item.update(**request.validated.get('body'))
    return item.to_dict()


@another_collection_service.post(validators=(service_post_validator,),
                                 schema=AnotherSchema)
def another_service_post(request):
    """ body_validator + schema
    """
    registry = request.anyblok.registry
    model = registry.get('Model.Example')
    item = model.insert(**request.validated['body'])
    return item.to_dict()


# thing endpoint through a service with the same model
thing_service = Service(name='thing_service', path='/things/{uuid}')

# thing collection endpoint through a service with the same model
thing_collection_service = Service(
    name='thing_collection_service',
    path='/things')


@thing_service.get(validators=(service_get_validator,),
                   schema=ThingRequestSchema)
def thing_service_get(request):
    """ Use Full validator with a request schema but Validate
    only the path
    """
    schema = ThingSchema()
    item = get_item(request, 'Model.Thing')
    return schema.dump(item)


@thing_service.put(validators=(service_put_validator,),
                   schema=ThingRequestSchema)
def thing_service_put(request):
    """ full_validator + RequestSchema
    """
    item = put_item(request, 'Model.Thing')
    schema = ThingSchema()
    return schema.dump(item)


@thing_service.delete(validators=(service_delete_validator,),
                      schema=ThingRequestSchema)
def thing_service_delete(request):
    """ full_validator + AnotherSchema
    """
    item = delete_item(request, 'Model.Thing')
    return item


@thing_collection_service.post(validators=(service_post_validator,),
                               schema=ThingRequestSchema)
def thing_service_post(request):
    """ body_validator + schema
    """
    item = post_item(request, 'Model.Thing')
    schema = ThingSchema()
    return schema.dump(item)


@thing_collection_service.get(validators=(service_collection_get_validator,),
                              schema=ThingRequestSchema)
def thing_service_get_collection(request):
    """ full_validator + querystring schema validation + schema validation
    for response data
    """
    schema = ThingSchema(many=True)
    collection = get_items(request, 'Model.Thing')
    return schema.dump(collection)


@resource(collection_path='/examples/with/errors',
          path='/examples/with/error/{id}',
          installed_blok=current_blok())
class ExampleResourceWithError(CrudResource):
    model = 'Model.Example'

    def add_error(self):
        self.request.errors.add('body', '500 Internal Server Error', 'test')
        self.request.errors.status = 500

    def create(self, Model, params):
        item = super(ExampleResourceWithError, self).create(Model, params)
        self.add_error()
        return item

    def update(self, item, params=None):
        super(ExampleResourceWithError, self).update(item, params=params)
        self.add_error()

    def delete_entry(self, item):
        super(ExampleResourceWithError, self).delete_entry(item)
        self.add_error()

    def collection_update(self, querystring, params=None):
        items = super(ExampleResourceWithError, self).collection_update(
            querystring, params=params)
        self.add_error()
        return items

    def delete_entries(self, querystring):
        count = super(ExampleResourceWithError, self).delete_entries(
            querystring)
        self.add_error()
        return count
